# -*- coding: utf-8 -*-
"""
py-iCSD toolbox!
Translation of the core functionality of the CSDplotter MATLAB package
to python.

The methods were originally developed by Klas H. Pettersen, as described in:
Klas H. Pettersen, Anna Devor, Istvan Ulbert, Anders M. Dale, Gaute T. Einevoll,
Current-source density estimation based on inversion of electrostatic forward
solution: Effects of finite extent of neuronal activity and conductivity
discontinuities, Journal of Neuroscience Methods, Volume 154, Issues 1-2,
30 June 2006, Pages 116-133, ISSN 0165-0270,
http://dx.doi.org/10.1016/j.jneumeth.2005.12.005.
(http://www.sciencedirect.com/science/article/pii/S0165027005004541)

The method themselves are implemented as callable subclasses of the base
CSD class object, which sets some common attributes,
and a basic function for calculating the iCSD, and a generic spatial filter
implementation.

The raw- and filtered CSD estimates are returned as Quantity arrays.

Requires pylab environment to work, i.e numpy+scipy+matplotlib, with the
addition of quantities (http://pythonhosted.org/quantities) and
neo (https://pythonhosted.org/neo)-

Original implementation from CSDplotter-0.1.1
(http://software.incf.org/software/csdplotter) by Klas. H. Pettersen 2005.

Written by:
- Espen.Hagen@umb.no, 2010,
- e.hagen@fz-juelich.de, 2015-2016

"""

import numpy as np
import scipy.integrate as si
import scipy.signal as ss
import quantities as pq


class CSD(object):
    """Base iCSD class"""
    def __init__(self, lfp, f_type='gaussian', f_order=(3, 1)):
        """Initialize parent class iCSD

        Parameters
        ----------
        lfp : np.ndarray * quantity.Quantity
            LFP signal of shape (# channels, # time steps)
        f_type : str
            type of spatial filter, must be a scipy.signal filter design method
        f_order : list
            settings for spatial filter, arg passed to  filter design function
        """
        self.name = 'CSD estimate parent class'
        self.lfp = lfp
        self.f_matrix = np.eye(lfp.shape[0]) * pq.m**3 / pq.S
        self.f_type = f_type
        self.f_order = f_order

    def get_csd(self, ):
        """
        Perform the CSD estimate from the LFP and forward matrix F, i.e as
        CSD=F**-1*LFP

        Arguments
        ---------

        Returns
        -------
        csd : np.ndarray * quantity.Quantity
            Array with the csd estimate
        """
        csd = np.linalg.solve(self.f_matrix, self.lfp)

        return csd * (self.f_matrix.units**-1 * self.lfp.units).simplified

    def filter_csd(self, csd, filterfunction='convolve'):
        """
        Spatial filtering of the CSD estimate, using an N-point filter

        Arguments
        ---------
        csd : np.ndarrray * quantity.Quantity
            Array with the csd estimate
        filterfunction : str
            'filtfilt' or 'convolve'. Apply spatial filter using
            scipy.signal.filtfilt or scipy.signal.convolve.
        """
        if self.f_type == 'gaussian':
            try:
                assert(len(self.f_order) == 2)
            except AssertionError as ae:
                raise ae('filter order f_order must be a tuple of length 2')
        else:
            try:
                assert(self.f_order > 0 and isinstance(self.f_order, int))
            except AssertionError as ae:
                raise ae('Filter order must be int > 0!')
        try:
            assert(filterfunction in ['filtfilt', 'convolve'])
        except AssertionError as ae:
            raise ae("{} not equal to 'filtfilt' or \
                     'convolve'".format(filterfunction))

        if self.f_type == 'boxcar':
            num = ss.boxcar(self.f_order)
            denom = np.array([num.sum()])
        elif self.f_type == 'hamming':
            num = ss.hamming(self.f_order)
            denom = np.array([num.sum()])
        elif self.f_type == 'triangular':
            num = ss.triang(self.f_order)
            denom = np.array([num.sum()])
        elif self.f_type == 'gaussian':
            num = ss.gaussian(self.f_order[0], self.f_order[1])
            denom = np.array([num.sum()])
        elif self.f_type == 'identity':
            num = np.array([1.])
            denom = np.array([1.])
        else:
            print('%s Wrong filter type!' % self.f_type)
            raise

        num_string = '[ '
        for i in num:
            num_string = num_string + '%.3f ' % i
        num_string = num_string + ']'
        denom_string = '[ '
        for i in denom:
            denom_string = denom_string + '%.3f ' % i
        denom_string = denom_string + ']'

        print(('discrete filter coefficients: \nb = {}, \
               \na = {}'.format(num_string, denom_string)))

        if filterfunction == 'filtfilt':
            return ss.filtfilt(num, denom, csd, axis=0) * csd.units
        elif filterfunction == 'convolve':
            csdf = csd / csd.units
            for i in range(csdf.shape[1]):
                csdf[:, i] = ss.convolve(csdf[:, i], num / denom.sum(), 'same')
            return csdf * csd.units


class StandardCSD(CSD):
    """
    Standard CSD method with and without Vaknin electrodes
    """

    def __init__(self, lfp, coord_electrode, **kwargs):
        """
        Initialize standard CSD method class with & without Vaknin electrodes.

        Parameters
        ----------
        lfp : np.ndarray * quantity.Quantity
            LFP signal of shape (# channels, # time steps) in units of V
        coord_electrode : np.ndarray * quantity.Quantity
            depth of evenly spaced electrode contact points of shape
            (# contacts, ) in units of m, must be monotonously increasing
        sigma : float * quantity.Quantity
            conductivity of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S/m
        vaknin_el : bool
            flag for using method of Vaknin to endpoint electrodes
            Defaults to True
        f_type : str
            type of spatial filter, must be a scipy.signal filter design method
            Defaults to 'gaussian'
        f_order : list
            settings for spatial filter, arg passed to  filter design function
            Defaults to (3,1) for the gaussian
        """
        self.parameters(**kwargs)
        CSD.__init__(self, lfp, self.f_type, self.f_order)

        diff_diff_coord = np.diff(np.diff(coord_electrode)).magnitude
        zeros_ddc = np.zeros_like(diff_diff_coord)
        try:
            assert(np.all(np.isclose(diff_diff_coord, zeros_ddc, atol=1e-12)))
        except AssertionError as ae:
            print('coord_electrode not monotonously varying')
            raise ae

        if self.vaknin_el:
            # extend lfps array by duplicating potential at endpoint contacts
            if lfp.ndim == 1:
                self.lfp = np.empty((lfp.shape[0] + 2,))
            else:
                self.lfp = np.empty((lfp.shape[0] + 2, lfp.shape[1]))
            self.lfp[0,] = lfp[0,]
            self.lfp[1:-1, ] = lfp
            self.lfp[-1,] = lfp[-1,]
            self.lfp = self.lfp * lfp.units
        else:
            self.lfp = lfp

        self.name = 'Standard CSD method'
        self.coord_electrode = coord_electrode

        self.f_inv_matrix = self.get_f_inv_matrix()

    def parameters(self, **kwargs):
        """Defining the default values of the method passed as kwargs
        Parameters
        ----------
        **kwargs
            Same as those passed to initialize the Class
        """
        self.sigma = kwargs.pop('sigma', 0.3 * pq.S / pq.m)
        self.vaknin_el = kwargs.pop('vaknin_el', True)
        self.f_type = kwargs.pop('f_type', 'gaussian')
        self.f_order = kwargs.pop('f_order', (3, 1))
        if kwargs:
            raise TypeError('Invalid keyword arguments:', kwargs.keys())

    def get_f_inv_matrix(self):
        """Calculate the inverse F-matrix for the standard CSD method"""
        h_val = abs(np.diff(self.coord_electrode)[0])
        f_inv = -np.eye(self.lfp.shape[0])

        # Inner matrix elements  is just the discrete laplacian coefficients
        for j in range(1, f_inv.shape[0] - 1):
            f_inv[j, j - 1: j + 2] = np.array([1., -2., 1.])
        return f_inv * -self.sigma / h_val

    def get_csd(self):
        """
        Perform the iCSD calculation, i.e: iCSD=F_inv*LFP

        Returns
        -------
        csd : np.ndarray * quantity.Quantity
            Array with the csd estimate
        """
        csd = np.dot(self.f_inv_matrix, self.lfp)[1:-1, ]
        # `np.dot()` does not return correct units, so the units of `csd` must
        # be assigned manually
        csd_units = (self.f_inv_matrix.units * self.lfp.units).simplified
        csd = csd.magnitude * csd_units

        return csd


class DeltaiCSD(CSD):
    """
    delta-iCSD method
    """
    def __init__(self, lfp, coord_electrode, **kwargs):
        """
        Initialize the delta-iCSD method class object

        Parameters
        ----------
        lfp : np.ndarray * quantity.Quantity
            LFP signal of shape (# channels, # time steps) in units of V
        coord_electrode : np.ndarray * quantity.Quantity
            depth of evenly spaced electrode contact points of shape
            (# contacts, ) in units of m
        diam : float * quantity.Quantity
            diamater of the assumed circular planar current sources centered
            at each contact
            Defaults to 500E-6 meters
        sigma : float * quantity.Quantity
            conductivity of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S / m
        sigma_top : float * quantity.Quantity
            conductivity on top of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S / m
        f_type : str
            type of spatial filter, must be a scipy.signal filter design method
            Defaults to 'gaussian'
        f_order : list
            settings for spatial filter, arg passed to  filter design function
            Defaults to (3,1) for gaussian
        """
        self.parameters(**kwargs)
        CSD.__init__(self, lfp, self.f_type, self.f_order)

        try:  # Should the class not take care of this?!
            assert(self.diam.units == coord_electrode.units)
        except AssertionError as ae:
            print('units of coord_electrode ({}) and diam ({}) differ'
                  .format(coord_electrode.units, self.diam.units))
            raise ae

        try:
            assert(np.all(np.diff(coord_electrode) > 0))
        except AssertionError as ae:
            print('values of coord_electrode not continously increasing')
            raise ae

        try:
            assert(self.diam.size == 1 or self.diam.size == coord_electrode.size)
            if self.diam.size == coord_electrode.size:
                assert(np.all(self.diam > 0 * self.diam.units))
            else:
                assert(self.diam > 0 * self.diam.units)
        except AssertionError as ae:
            print('diam must be positive scalar or of same shape \
                   as coord_electrode')
            raise ae
        if self.diam.size == 1:
            self.diam = np.ones(coord_electrode.size) * self.diam

        self.name = 'delta-iCSD method'
        self.coord_electrode = coord_electrode

        # initialize F- and iCSD-matrices
        self.f_matrix = np.empty((self.coord_electrode.size,
                                  self.coord_electrode.size))
        self.f_matrix = self.get_f_matrix()

    def parameters(self, **kwargs):
        """Defining the default values of the method passed as kwargs
        Parameters
        ----------
        **kwargs
            Same as those passed to initialize the Class
        """
        self.diam = kwargs.pop('diam', 500E-6 * pq.m)
        self.sigma = kwargs.pop('sigma', 0.3 * pq.S / pq.m)
        self.sigma_top = kwargs.pop('sigma_top', 0.3 * pq.S / pq.m)
        self.f_type = kwargs.pop('f_type', 'gaussian')
        self.f_order = kwargs.pop('f_order', (3, 1))
        if kwargs:
            raise TypeError('Invalid keyword arguments:', kwargs.keys())

    def get_f_matrix(self):
        """Calculate the F-matrix"""
        f_matrix = np.empty((self.coord_electrode.size,
                             self.coord_electrode.size)) * self.coord_electrode.units
        for j in range(self.coord_electrode.size):
            for i in range(self.coord_electrode.size):
                f_matrix[j, i] = ((np.sqrt((self.coord_electrode[j] -
                                            self.coord_electrode[i])**2 +
                    (self.diam[j] / 2)**2) - abs(self.coord_electrode[j] -
                                                 self.coord_electrode[i])) +
                    (self.sigma - self.sigma_top) / (self.sigma +
                                                     self.sigma_top) *
                    (np.sqrt((self.coord_electrode[j] +
                              self.coord_electrode[i])**2 + (self.diam[j] / 2)**2)-
                    abs(self.coord_electrode[j] + self.coord_electrode[i])))

        f_matrix /= (2 * self.sigma)
        return f_matrix


class StepiCSD(CSD):
    """step-iCSD method"""
    def __init__(self, lfp, coord_electrode, **kwargs):

        """
        Initializing step-iCSD method class object

        Parameters
        ----------
        lfp : np.ndarray * quantity.Quantity
            LFP signal of shape (# channels, # time steps) in units of V
        coord_electrode : np.ndarray * quantity.Quantity
            depth of evenly spaced electrode contact points of shape
            (# contacts, ) in units of m
        diam : float or np.ndarray * quantity.Quantity
            diameter(s) of the assumed circular planar current sources centered
            at each contact
            Defaults to 500E-6 meters
        h : float or np.ndarray * quantity.Quantity
            assumed thickness of the source cylinders at all or each contact
            Defaults to np.ones(15) * 100E-6 * pq.m
        sigma : float * quantity.Quantity
            conductivity of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S / m
        sigma_top : float * quantity.Quantity
            conductivity on top of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S / m
        tol : float
            tolerance of numerical integration
            Defaults 1e-6
        f_type : str
            type of spatial filter, must be a scipy.signal filter design method
            Defaults to 'gaussian'
        f_order : list
            settings for spatial filter, arg passed to  filter design function
            Defaults to (3,1) for the gaussian
        """
        self.parameters(**kwargs)
        CSD.__init__(self, lfp, self.f_type, self.f_order)

        try:  # Should the class not take care of this?
            assert(self.diam.units == coord_electrode.units)
        except AssertionError as ae:
            print('units of coord_electrode ({}) and diam ({}) differ'
                  .format(coord_electrode.units, self.diam.units))
            raise ae
        try:
            assert(np.all(np.diff(coord_electrode) > 0))
        except AssertionError as ae:
            print('values of coord_electrode not continously increasing')
            raise ae

        try:
            assert(self.diam.size == 1 or self.diam.size == coord_electrode.size)
            if self.diam.size == coord_electrode.size:
                assert(np.all(self.diam > 0 * self.diam.units))
            else:
                assert(self.diam > 0 * self.diam.units)
        except AssertionError as ae:
            print('diam must be positive scalar or of same shape \
                   as coord_electrode')
            raise ae
        if self.diam.size == 1:
            self.diam = np.ones(coord_electrode.size) * self.diam
        try:
            assert(self.h.size == 1 or self.h.size == coord_electrode.size)
            if self.h.size == coord_electrode.size:
                assert(np.all(self.h > 0 * self.h.units))
        except AssertionError as ae:
            print('h must be scalar or of same shape as coord_electrode')
            raise ae
        if self.h.size == 1:
            self.h = np.ones(coord_electrode.size) * self.h

        self.name = 'step-iCSD method'
        self.coord_electrode = coord_electrode

        # compute forward-solution matrix
        self.f_matrix = self.get_f_matrix()

    def parameters(self, **kwargs):
        """Defining the default values of the method passed as kwargs
        Parameters
        ----------
        **kwargs
            Same as those passed to initialize the Class
        """

        self.diam = kwargs.pop('diam', 500E-6 * pq.m)
        self.h = kwargs.pop('h', np.ones(23) * 100E-6 * pq.m)
        self.sigma = kwargs.pop('sigma', 0.3 * pq.S / pq.m)
        self.sigma_top = kwargs.pop('sigma_top', 0.3 * pq.S / pq.m)
        self.tol = kwargs.pop('tol', 1e-6)
        self.f_type = kwargs.pop('f_type', 'gaussian')
        self.f_order = kwargs.pop('f_order', (3, 1))
        if kwargs:
            raise TypeError('Invalid keyword arguments:', kwargs.keys())

    def get_f_matrix(self):
        """Calculate F-matrix for step iCSD method"""
        el_len = self.coord_electrode.size
        f_matrix = np.zeros((el_len, el_len))
        for j in range(el_len):
            for i in range(el_len):
                lower_int = self.coord_electrode[i] - self.h[j] / 2
                if lower_int < 0:
                    lower_int = self.h[j].units
                upper_int = self.coord_electrode[i] + self.h[j] / 2

                # components of f_matrix object
                f_cyl0 = si.quad(self._f_cylinder,
                                 a=lower_int, b=upper_int,
                                 args=(float(self.coord_electrode[j]),
                                       float(self.diam[j]),
                                       float(self.sigma)),
                                 epsabs=self.tol)[0]
                f_cyl1 = si.quad(self._f_cylinder, a=lower_int, b=upper_int,
                                 args=(-float(self.coord_electrode[j]),
                                       float(self.diam[j]), float(self.sigma)),
                                 epsabs=self.tol)[0]

                # method of images coefficient
                mom = (self.sigma - self.sigma_top) / (self.sigma + self.sigma_top)

                f_matrix[j, i] = f_cyl0 + mom * f_cyl1

        # assume si.quad trash the units
        return f_matrix * self.h.units**2 / self.sigma.units

    def _f_cylinder(self, zeta, z_val, diam, sigma):
        """function used by class method"""
        f_cyl = 1. / (2. * sigma) * \
            (np.sqrt((diam / 2)**2 + ((z_val - zeta))**2) - abs(z_val - zeta))
        return f_cyl


class SplineiCSD(CSD):
    """spline iCSD method"""
    def __init__(self, lfp, coord_electrode, **kwargs):

        """
        Initializing spline-iCSD method class object

        Parameters
        ----------
        lfp : np.ndarray * quantity.Quantity
            LFP signal of shape (# channels, # time steps) in units of V
        coord_electrode : np.ndarray * quantity.Quantity
            depth of evenly spaced electrode contact points of shape
            (# contacts, ) in units of m
        diam : float * quantity.Quantity
            diamater of the assumed circular planar current sources centered
            at each contact
            Defaults to 500E-6 meters
        sigma : float * quantity.Quantity
            conductivity of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S / m
        sigma_top : float * quantity.Quantity
            conductivity on top of tissue in units of S/m or 1/(ohm*m)
            Defaults to 0.3 S / m
        tol : float
            tolerance of numerical integration
            Defaults 1e-6
        f_type : str
            type of spatial filter, must be a scipy.signal filter design method
            Defaults to 'gaussian'
        f_order : list
            settings for spatial filter, arg passed to  filter design function
            Defaults to (3,1) for the gaussian
        num_steps : int
            number of data points for the spatially upsampled LFP/CSD data
            Defaults to 200
        """
        self.parameters(**kwargs)
        CSD.__init__(self, lfp, self.f_type, self.f_order)

        try:  # Should the class not take care of this?!
            assert(self.diam.units == coord_electrode.units)
        except AssertionError as ae:
            print('units of coord_electrode ({}) and diam ({}) differ'
                  .format(coord_electrode.units, self.diam.units))
            raise
        try:
            assert(np.all(np.diff(coord_electrode) > 0))
        except AssertionError as ae:
            print('values of coord_electrode not continously increasing')
            raise ae

        try:
            assert(self.diam.size == 1 or self.diam.size == coord_electrode.size)
            if self.diam.size == coord_electrode.size:
                assert(np.all(self.diam > 0 * self.diam.units))
        except AssertionError as ae:
            print('diam must be scalar or of same shape as coord_electrode')
            raise ae
        if self.diam.size == 1:
            self.diam = np.ones(coord_electrode.size) * self.diam

        self.name = 'spline-iCSD method'
        self.coord_electrode = coord_electrode

        # compute stuff
        self.f_matrix = self.get_f_matrix()

    def parameters(self, **kwargs):
        """Defining the default values of the method passed as kwargs
        Parameters
        ----------
        **kwargs
            Same as those passed to initialize the Class
        """
        self.diam = kwargs.pop('diam', 500E-6 * pq.m)
        self.sigma = kwargs.pop('sigma', 0.3 * pq.S / pq.m)
        self.sigma_top = kwargs.pop('sigma_top', 0.3 * pq.S / pq.m)
        self.tol = kwargs.pop('tol', 1e-6)
        self.num_steps = kwargs.pop('num_steps', 200)
        self.f_type = kwargs.pop('f_type', 'gaussian')
        self.f_order = kwargs.pop('f_order', (3, 1))
        if kwargs:
            raise TypeError('Invalid keyword arguments:', kwargs.keys())

    def get_f_matrix(self):
        """Calculate the F-matrix for cubic spline iCSD method"""
        el_len = self.coord_electrode.size
        z_js = np.zeros(el_len + 1)
        z_js[:-1] = np.array(self.coord_electrode)
        z_js[-1] = z_js[-2] + float(np.diff(self.coord_electrode).mean())

        # Define integration matrixes
        f_mat0 = np.zeros((el_len, el_len + 1))
        f_mat1 = np.zeros((el_len, el_len + 1))
        f_mat2 = np.zeros((el_len, el_len + 1))
        f_mat3 = np.zeros((el_len, el_len + 1))

        # Calc. elements
        for j in range(el_len):
            for i in range(el_len):
                f_mat0[j, i] = si.quad(self._f_mat0, a=z_js[i], b=z_js[i + 1],
                                       args=(z_js[j + 1],
                                             float(self.sigma),
                                             float(self.diam[j])),
                                       epsabs=self.tol)[0]
                f_mat1[j, i] = si.quad(self._f_mat1, a=z_js[i], b=z_js[i + 1],
                                       args=(z_js[j + 1], z_js[i],
                                             float(self.sigma),
                                             float(self.diam[j])),
                                       epsabs=self.tol)[0]
                f_mat2[j, i] = si.quad(self._f_mat2, a=z_js[i], b=z_js[i + 1],
                                       args=(z_js[j + 1], z_js[i],
                                             float(self.sigma),
                                             float(self.diam[j])),
                                       epsabs=self.tol)[0]
                f_mat3[j, i] = si.quad(self._f_mat3, a=z_js[i], b=z_js[i + 1],
                                       args=(z_js[j + 1], z_js[i],
                                             float(self.sigma),
                                             float(self.diam[j])),
                                       epsabs=self.tol)[0]

                # image technique if conductivity not constant:
                if self.sigma != self.sigma_top:
                    f_mat0[j, i] = f_mat0[j, i] + (self.sigma-self.sigma_top) / \
                                                (self.sigma + self.sigma_top) * \
                            si.quad(self._f_mat0, a=z_js[i], b=z_js[i+1], \
                                    args=(-z_js[j+1],
                                          float(self.sigma), float(self.diam[j])), \
                                    epsabs=self.tol)[0]
                    f_mat1[j, i] = f_mat1[j, i] + (self.sigma-self.sigma_top) / \
                        (self.sigma + self.sigma_top) * \
                            si.quad(self._f_mat1, a=z_js[i], b=z_js[i+1], \
                                args=(-z_js[j+1], z_js[i], float(self.sigma),
                                      float(self.diam[j])), epsabs=self.tol)[0]
                    f_mat2[j, i] = f_mat2[j, i] + (self.sigma-self.sigma_top) / \
                        (self.sigma + self.sigma_top) * \
                            si.quad(self._f_mat2, a=z_js[i], b=z_js[i+1], \
                                args=(-z_js[j+1], z_js[i], float(self.sigma),
                                      float(self.diam[j])), epsabs=self.tol)[0]
                    f_mat3[j, i] = f_mat3[j, i] + (self.sigma-self.sigma_top) / \
                        (self.sigma + self.sigma_top) * \
                            si.quad(self._f_mat3, a=z_js[i], b=z_js[i+1], \
                                args=(-z_js[j+1], z_js[i], float(self.sigma),
                                      float(self.diam[j])), epsabs=self.tol)[0]

        e_mat0, e_mat1, e_mat2, e_mat3 = self._calc_e_matrices()

        # Calculate the F-matrix
        f_matrix = np.eye(el_len + 2)
        f_matrix[1:-1, :] = np.dot(f_mat0, e_mat0) + \
                            np.dot(f_mat1, e_mat1) + \
                            np.dot(f_mat2, e_mat2) + \
                            np.dot(f_mat3, e_mat3)

        return f_matrix * self.coord_electrode.units**2 / self.sigma.units

    def get_csd(self):
        """
        Calculate the iCSD using the spline iCSD method

        Returns
        -------
        csd : np.ndarray * quantity.Quantity
            Array with csd estimate


        """
        e_mat = self._calc_e_matrices()

        el_len = self.coord_electrode.size

        # padding the lfp with zeros on top/bottom
        if self.lfp.ndim == 1:
            cs_lfp = np.r_[[0], np.asarray(self.lfp), [0]].reshape(1, -1).T
            csd = np.zeros(self.num_steps)
        else:
            cs_lfp = np.vstack((np.zeros(self.lfp.shape[1]),
                                np.asarray(self.lfp),
                                np.zeros(self.lfp.shape[1])))
            csd = np.zeros((self.num_steps, self.lfp.shape[1]))
        cs_lfp *= self.lfp.units

        # CSD coefficients
        csd_coeff = np.linalg.solve(self.f_matrix, cs_lfp)

        # The cubic spline polynomial coefficients
        a_mat0 = np.dot(e_mat[0], csd_coeff)
        a_mat1 = np.dot(e_mat[1], csd_coeff)
        a_mat2 = np.dot(e_mat[2], csd_coeff)
        a_mat3 = np.dot(e_mat[3], csd_coeff)

        # Extend electrode coordinates in both end by min contact interdistance
        h = np.diff(self.coord_electrode).min()
        z_js = np.zeros(el_len + 2)
        z_js[0] = self.coord_electrode[0] - h
        z_js[1: -1] = self.coord_electrode
        z_js[-1] = self.coord_electrode[-1] + h

        # create high res spatial grid
        out_zs = np.linspace(z_js[1], z_js[-2], self.num_steps)

        # Calculate iCSD estimate on grid from polynomial coefficients.
        i = 0
        for j in range(self.num_steps):
            if out_zs[j] >= z_js[i + 1]:
                i += 1
            csd[j] = (a_mat0[i, :] + a_mat1[i, :] *
                      (out_zs[j] - z_js[i]) +
                      a_mat2[i, :] * (out_zs[j] - z_js[i])**2 +
                      a_mat3[i, :] * (out_zs[j] - z_js[i])**3).item()

        csd_unit = (self.f_matrix.units**-1 * self.lfp.units).simplified

        return csd * csd_unit

    def _f_mat0(self, zeta, z_val, sigma, diam):
        """0'th order potential function"""
        return 1. / (2. * sigma) * \
            (np.sqrt((diam / 2)**2 + ((z_val - zeta))**2) - abs(z_val - zeta))

    def _f_mat1(self, zeta, z_val, zi_val, sigma, diam):
        """1'th order potential function"""
        return (zeta - zi_val) * self._f_mat0(zeta, z_val, sigma, diam)

    def _f_mat2(self, zeta, z_val, zi_val, sigma, diam):
        """2'nd order potential function"""
        return (zeta - zi_val)**2 * self._f_mat0(zeta, z_val, sigma, diam)

    def _f_mat3(self, zeta, z_val, zi_val, sigma, diam):
        """3'rd order potential function"""
        return (zeta - zi_val)**3 * self._f_mat0(zeta, z_val, sigma, diam)

    def _calc_k_matrix(self):
        """Calculate the K-matrix used by to calculate E-matrices"""
        el_len = self.coord_electrode.size
        h = float(np.diff(self.coord_electrode).min())

        c_jm1 = np.eye(el_len + 2, k=0) / h
        c_jm1[0, 0] = 0

        c_j0 = np.eye(el_len + 2) / h
        c_j0[-1, -1] = 0

        c_jall = c_j0
        c_jall[0, 0] = 1
        c_jall[-1, -1] = 1

        tjp1 = np.eye(el_len + 2, k=1)
        tjm1 = np.eye(el_len + 2, k=-1)

        tj0 = np.eye(el_len + 2)
        tj0[0, 0] = 0
        tj0[-1, -1] = 0

        # Defining K-matrix used to calculate e_mat1-3
        return np.dot(np.linalg.inv(np.dot(c_jm1, tjm1) +
                                    2 * np.dot(c_jm1, tj0) +
                                    2 * c_jall +
                                    np.dot(c_j0, tjp1)),
                      3 * (np.dot(np.dot(c_jm1, c_jm1), tj0) -
                           np.dot(np.dot(c_jm1, c_jm1), tjm1) +
                           np.dot(np.dot(c_j0, c_j0), tjp1) -
                           np.dot(np.dot(c_j0, c_j0), tj0)))

    def _calc_e_matrices(self):
        """Calculate the E-matrices used by cubic spline iCSD method"""
        el_len = self.coord_electrode.size
        # expanding electrode grid
        h = float(np.diff(self.coord_electrode).min())

        # Define transformation matrices
        c_mat3 = np.eye(el_len + 1) / h

        # Get K-matrix
        k_matrix = self._calc_k_matrix()

        # Define matrixes for C to A transformation:
        tja = np.eye(el_len + 2)[:-1, ]
        tjp1a = np.eye(el_len + 2, k=1)[:-1, ]

        # Define spline coefficients
        e_mat0 = tja
        e_mat1 = np.dot(tja, k_matrix)
        e_mat2 = 3 * np.dot(c_mat3**2, (tjp1a - tja)) - \
                            np.dot(np.dot(c_mat3, (tjp1a + 2 * tja)), k_matrix)
        e_mat3 = 2 * np.dot(c_mat3**3, (tja - tjp1a)) + \
                            np.dot(np.dot(c_mat3**2, (tjp1a + tja)), k_matrix)

        return e_mat0, e_mat1, e_mat2, e_mat3


if __name__ == '__main__':
    from scipy.io import loadmat
    import matplotlib.pyplot as plt

    
    #loading test data
    test_data = loadmat('test_data.mat')
    
    #prepare lfp data for use, by changing the units to SI and append quantities,
    #along with electrode geometry, conductivities and assumed source geometry
    lfp_data = test_data['pot1'] * 1E-6 * pq.V        # [uV] -> [V]
    z_data = np.linspace(100E-6, 2300E-6, 23) * pq.m  # [m]
    diam = 500E-6 * pq.m                              # [m]
    h = 100E-6 * pq.m                                 # [m]
    sigma = 0.3 * pq.S / pq.m                         # [S/m] or [1/(ohm*m)]
    sigma_top = 0.3 * pq.S / pq.m                     # [S/m] or [1/(ohm*m)]
    
    # Input dictionaries for each method
    delta_input = {
        'lfp' : lfp_data,
        'coord_electrode' : z_data,
        'diam' : diam,          # source diameter
        'sigma' : sigma,        # extracellular conductivity
        'sigma_top' : sigma,    # conductivity on top of cortex
        'f_type' : 'gaussian',  # gaussian filter
        'f_order' : (3, 1),     # 3-point filter, sigma = 1.
    }
    step_input = {
        'lfp' : lfp_data,
        'coord_electrode' : z_data,
        'diam' : diam,
        'h' : h,                # source thickness
        'sigma' : sigma,
        'sigma_top' : sigma,
        'tol' : 1E-12,          # Tolerance in numerical integration
        'f_type' : 'gaussian',
        'f_order' : (3, 1),
    }
    spline_input = {
        'lfp' : lfp_data,
        'coord_electrode' : z_data,
        'diam' : diam,
        'sigma' : sigma,
        'sigma_top' : sigma,
        'num_steps' : 201,      # Spatial CSD upsampling to N steps
        'tol' : 1E-12,
        'f_type' : 'gaussian',
        'f_order' : (20, 5),
    }
    std_input = {
        'lfp' : lfp_data,
        'coord_electrode' : z_data,
        'sigma' : sigma,
        'f_type' : 'gaussian',
        'f_order' : (3, 1),
    }
    
    
    #Create the different CSD-method class instances. We use the class methods
    #get_csd() and filter_csd() below to get the raw and spatially filtered
    #versions of the current-source density estimates.
    csd_dict = dict(
        delta_icsd = DeltaiCSD(**delta_input),
        step_icsd = StepiCSD(**step_input),
        spline_icsd = SplineiCSD(**spline_input),
        std_csd = StandardCSD(**std_input),
    )
    
    #plot
    for method, csd_obj in list(csd_dict.items()):
        fig, axes = plt.subplots(3,1, figsize=(8,8))
    
        #plot LFP signal
        ax = axes[0]
        im = ax.imshow(np.array(lfp_data), origin='upper', vmin=-abs(lfp_data).max(), \
                  vmax=abs(lfp_data).max(), cmap='jet_r', interpolation='nearest')
        ax.axis(ax.axis('tight'))
        cb = plt.colorbar(im, ax=ax)
        cb.set_label('LFP (%s)' % lfp_data.dimensionality.string)
        ax.set_xticklabels([])
        ax.set_title('LFP')
        ax.set_ylabel('ch #')
    
        #plot raw csd estimate
        csd = csd_obj.get_csd()
        ax = axes[1]
        im = ax.imshow(np.array(csd), origin='upper', vmin=-abs(csd).max(), \
              vmax=abs(csd).max(), cmap='jet_r', interpolation='nearest')
        ax.axis(ax.axis('tight'))
        ax.set_title(csd_obj.name)
        cb = plt.colorbar(im, ax=ax)
        cb.set_label('CSD (%s)' % csd.dimensionality.string)
        ax.set_xticklabels([])
        ax.set_ylabel('ch #')
    
        #plot spatially filtered csd estimate
        ax = axes[2]
        csd = csd_obj.filter_csd(csd)
        im = ax.imshow(np.array(csd), origin='upper', vmin=-abs(csd).max(), \
              vmax=abs(csd).max(), cmap='jet_r', interpolation='nearest')
        ax.axis(ax.axis('tight'))
        ax.set_title(csd_obj.name + ', filtered')
        cb = plt.colorbar(im, ax=ax)
        cb.set_label('CSD (%s)' % csd.dimensionality.string)
        ax.set_ylabel('ch #')
        ax.set_xlabel('timestep')
    
    
    plt.show()




"""
This module provides an implementation of the formal concept analysis (FCA) in
pure Python. It is used by the SPADE analysis. The code builds on C. Lindig's
Fast Concept Analysis work (1999,2002).

Original code available at:

Copyright (C) 2008-2019 by Dominik Endres (dominik.endres@gmail.com).
Relicensed for Elephant by permission.

Usage example:
--------------

>>> relation=[]
>>> relation+=[('monkeyHand','neuron2')]
>>> relation+=[('monkeyFace','neuron1')]
>>> relation+=[('monkeyFace','neuron2')]
>>> relation+=[('humanFace','neuron1')]
>>> relation+=[('spider','neuron3')]
>>> concepts=FormalConcepts(relation)
>>> concepts.computeLattice()

If you generate publications based on this code, please cite the following
paper:

Endres D., Adam R., Giese M.A. & Noppeney U.. (2012).
Understanding the Semantic Structure of Human fMRI Brain Recordings with Formal
Concept Analysis. Proceedings of the 10th International Conference on Formal
Concept Analysis (ICFCA 2012) LNAI 7278, Springer,pp. 96-111.
"""

import bisect
from collections import defaultdict

import tqdm


class FormalConcept(object):
    """
    A formal concept is comprised of an extent and and intent.
    Furthermore, intentIndexes is an ordered list of attribute indexes for
    lectic ordering. Also contains sets of introduced attibutes and objects
    and lectically ordered lists of upper and lower neighbours.
    """

    def __init__(self, extent=frozenset(), intent=frozenset(),
                 intentIndexes=[]):
        """ intent/extent are a frozensets because they need to be hashable."""
        self.cnum = 0
        self.extent = extent
        self.intent = intent
        self.introducedAttributes = set()
        self.introducedObjects = set()
        self.intentIndexes = intentIndexes
        self.upperNeighbours = []
        self.lowerNeighbours = []
        self.visited = False  # for lattice traversal

        # attributes that were introduced closest in upwards direction
        # useful for naming a concept that introduces no attributes.
        # recompute after pruning!
        self.closestIntroducedAttributes = []
        # all attributes that are introduced in the downset of this concept;
        # useful for building search list.
        self.downsetAttributes = set()

    def __eq__(self, other):
        """lectic order on intentIndexes."""
        return self.intentIndexes == other.intentIndexes

    def __lt__(self, other):
        """lectic order on intentIndexes."""
        if self.intentIndexes == other.intentIndexes:
            return -1
        i1 = 0
        i2len = len(other.intentIndexes)
        for a1 in self.intentIndexes:
            if i1 >= i2len:
                return -1
            a2 = other.intentIndexes[i1]
            if a1 > a2:
                return 1
            elif a1 < a2:
                return -1
            i1 += 1
        return 1

    def __repr__(self):
        """ print the concept."""
        strrep = "concept no:" + str(self.cnum) + "\n"
        strrep += "extent:" + repr(self.extent) + "\n"
        strrep += "intent:" + repr(self.intent) + "\n"
        strrep += "introduced objects:" + repr(self.introducedObjects) + "\n"
        strrep += "introduced attributes:" + repr(
            self.introducedAttributes) + "\n"
        if hasattr(self, "stability"):
            strrep += "stability: {0:1.4f}".format(
                self.stability) + "\n"
        strrep += "upper neighbours: "
        for un in self.upperNeighbours:
            strrep += str(un.cnum) + ", "
        strrep += "\n"
        strrep += "lower neighbours: "
        for ln in self.lowerNeighbours:
            strrep += str(ln.cnum) + ", "
        strrep += "\n"
        return strrep


class FormalContext(object):
    """
    The formal context.
    Builds dictionaries object=>attributes and vice versa for faster closure
    computation.
    Set of objects and attributes are kept in lists rather than sets for lectic
    ordering of concepts.
    """

    def __init__(self, relation, objects=None, attributes=None):
        """
        'relation' has to be an iterable container of tuples.
        If objects or attributes are not supplied, determine from relation.
        """
        # map from object=> set of attributes of this object
        self.objectsToAttributes = defaultdict(set)
        # map from attributes => set of objects of this attribute
        self.attributesToObjects = defaultdict(set)
        # objects and attributes are kept in lists rather than sets for lectic
        # ordering of concepts.
        self.objects = set()
        self.attributes = set()
        if objects is not None:
            self.objects.update(objects)
        if attributes is not None:
            self.attributes.update(attributes)

        for obj, att in relation:
            self.objects.add(obj)
            self.attributes.add(att)
            self.objectsToAttributes[obj].add(att)
            self.attributesToObjects[att].add(obj)

        self.attributes = sorted(self.attributes, reverse=True)

    def objectsPrime(self, objectSet):
        """return a frozenset of all attributes which are shared by members
        of objectSet. """
        if len(objectSet) == 0:
            return frozenset(self.attributes)
        oiter = iter(objectSet)
        opr = self.objectsToAttributes[next(oiter)].copy()
        for obj in oiter:
            opr.intersection_update(self.objectsToAttributes[obj])
        return frozenset(opr)

    def attributesPrime(self, attributeSet):
        """return a set of all objects which have all attributes in
        attribute set. """
        if len(attributeSet) == 0:
            return frozenset(self.objects)
        aiter = iter(attributeSet)
        apr = self.attributesToObjects[next(aiter)].copy()
        for att in aiter:
            apr.intersection_update(self.attributesToObjects[att])
        return frozenset(apr)

    def updateIntent(self, intent, object):
        """return intersection of intent and all attributes of object."""
        return frozenset(intent.intersection(self.objectsToAttributes[object]))

    def indexList(self, attributeSet):
        """return ordered list of attribute indexes. For lectic ordering of
        concepts. """
        ilist = sorted(map(self.attributes.index, attributeSet))
        return ilist


class FormalConcepts(object):
    """ Computes set of concepts from a binary relation by an algorithm
    similar to C. Lindig's Fast Concept Analysis (2002).
    """

    def __init__(self, relation, objects=None, attributes=None, verbose=False):
        """ 'relation' has to be an iterable container of tuples. If objects
        or attributes are not supplied, determine from relation. """
        self.context = FormalContext(relation, objects, attributes)
        self.concepts = []  # a lectically ordered list of concepts"
        self.intentToConceptDict = dict()
        self.verbose = verbose

    def computeUpperNeighbours(self, concept):
        """
        This version of upperNeighbours runs fast enough in Python to be
        useful.
        Based on a theorem from C. Lindig's (1999) PhD thesis.
        Returns list of upper neighbours of concept."""
        # The set of all objects g which are not in concept's extent G and
        # might therefore be used to create upper neighbours via ((G u g)'',(G
        # u g)')
        upperNeighbourGeneratingObjects = self.context.objects.difference(
            concept.extent)
        # dictionary of intent => set of generating objects
        upperNeighbourCandidates = defaultdict(set)
        for g in upperNeighbourGeneratingObjects:
            # an intent of a concept >= concept. Computed by intersecting i(g)
            # with concept.intent,
            # where i(g) is the set of all attributes of g.
            intent = self.context.updateIntent(concept.intent, g)
            # self.intentToConceptDict is a dictionary of all concepts computed
            # so far.
            if intent not in self.intentToConceptDict:
                # Store every concept in self.conceptDict, because it will
                # eventually be used and the closure is expensive to compute
                extent = self.context.attributesPrime(intent)
                curConcept = FormalConcept(extent, intent,
                                           self.context.indexList(intent))
                self.intentToConceptDict[intent] = curConcept

            # remember which g generated what concept
            upperNeighbourCandidates[intent].add(g)

        neighbours = []
        # find all upper neighbours by Lindig's theorem:
        # a concept C=((G u g)'',(G u g)') is an upper neighbour of (G,I) if
        # and only if (G u g)'' \ G = set of all g which generated C.
        for intent, generatingObjects in upperNeighbourCandidates.items():
            extraObjects = self.intentToConceptDict[intent].extent.difference(
                concept.extent)
            if extraObjects == generatingObjects:
                neighbours.append(self.intentToConceptDict[intent])
        return neighbours

    def numberConceptsAndComputeIntroduced(self):
        """ Numbers concepts and computes introduced objects and attributes"""
        for curConNum, curConcept in enumerate(self.concepts):
            curConcept.cnum = curConNum
            curConcept.introducedObjects = set(curConcept.extent)
            for ln in curConcept.lowerNeighbours:
                curConcept.introducedObjects.difference_update(ln.extent)
            curConcept.introducedAttributes = set(curConcept.intent)
            for un in curConcept.upperNeighbours:
                curConcept.introducedAttributes.difference_update(un.intent)

    def computeLattice(self):
        """ Computes concepts and lattice.
        self.concepts contains lectically ordered list of concepts after
        completion."""
        intent = self.context.objectsPrime(set())
        extent = self.context.attributesPrime(intent)
        curConcept = FormalConcept(extent, intent,
                                   self.context.indexList(intent))
        self.concepts = [curConcept]
        self.intentToConceptDict[curConcept.intent] = curConcept
        curConceptIndex = 0
        progress_bar = tqdm.tqdm(disable=not self.verbose,
                                 desc="computeLattice")
        while curConceptIndex >= 0:
            upperNeighbours = self.computeUpperNeighbours(curConcept)
            for upperNeighbour in upperNeighbours:
                upperNeighbourIndex = bisect.bisect(self.concepts,
                                                    upperNeighbour)
                if upperNeighbourIndex == 0 or self.concepts[
                        upperNeighbourIndex - 1] != upperNeighbour:
                    self.concepts.insert(upperNeighbourIndex, upperNeighbour)
                    curConceptIndex += 1

                curConcept.upperNeighbours.append(upperNeighbour)
                upperNeighbour.lowerNeighbours.append(curConcept)

            curConceptIndex -= 1
            curConcept = self.concepts[curConceptIndex]
            progress_bar.update()

        self.numberConceptsAndComputeIntroduced()



# -*- coding: utf-8 -*-
"""
Tools to manipulate Neo objects.

.. autosummary::
    :toctree: _toctree/neo_tools

    extract_neo_attributes
    get_all_spiketrains
    get_all_events
    get_all_epochs

:copyright: Copyright 2014-2023 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals
import warnings

from itertools import chain

from neo.core.spiketrainlist import SpikeTrainList
from neo.core.container import unique_objs
from elephant.utils import deprecated_alias

__all__ = [
    "extract_neo_attributes",
    "get_all_spiketrains",
    "get_all_events",
    "get_all_epochs"
]


@deprecated_alias(obj='neo_object')
def extract_neo_attributes(neo_object, parents=True, child_first=True,
                           skip_array=False, skip_none=False):
    """
    Given a Neo object, return a dictionary of attributes and annotations.

    Parameters
    ----------
    neo_object : neo.BaseNeo
        Object to get attributes and annotations.
    parents : bool, optional
        If True, also include attributes and annotations from parent Neo
        objects (if any).
        Default: True
    child_first : bool, optional
        If True, values of child attributes are used over parent attributes in
        the event of a name conflict.
        If False, parent attributes are used.
        This parameter does nothing if `parents` is False.
        Default: True
    skip_array : bool, optional
        If True, skip attributes that store non-scalar array values.
        Default: False
    skip_none : bool, optional
        If True, skip annotations and attributes that have a value of None.
        Default: False

    Returns
    -------
    dict
        A dictionary where the keys are annotations or attribute names and
        the values are the corresponding annotation or attribute value.

    """
    attrs = neo_object.annotations.copy()
    if not skip_array and hasattr(neo_object, "array_annotations"):
        # Exclude labels and durations, and any other fields that should not
        # be a part of array_annotation.
        required_keys = set(neo_object.array_annotations).difference(
            dir(neo_object))
        for a in required_keys:
            if "array_annotations" not in attrs:
                attrs["array_annotations"] = {}
            attrs["array_annotations"][a] = \
                neo_object.array_annotations[a].copy()
    for attr in neo_object._necessary_attrs + neo_object._recommended_attrs:
        if skip_array and len(attr) >= 3 and attr[2]:
            continue
        attr = attr[0]
        if attr == getattr(neo_object, '_quantity_attr', None):
            continue
        attrs[attr] = getattr(neo_object, attr, None)

    if skip_none:
        for attr, value in attrs.copy().items():
            if value is None:
                del attrs[attr]

    if not parents:
        return attrs

    for parent in getattr(neo_object, 'parents', []):
        if parent is None:
            continue
        newattr = extract_neo_attributes(parent, parents=True,
                                         child_first=child_first,
                                         skip_array=skip_array,
                                         skip_none=skip_none)
        if child_first:
            newattr.update(attrs)
            attrs = newattr
        else:
            attrs.update(newattr)

    return attrs


def extract_neo_attrs(*args, **kwargs):
    warnings.warn("'extract_neo_attrs' function is deprecated; "
                  "use 'extract_neo_attributes'", DeprecationWarning)
    return extract_neo_attributes(*args, **kwargs)


def _get_all_objs(container, class_name):
    """
    Get all Neo objects of a given type from a container.

    The objects can be any list, dict, or other iterable or mapping containing
    Neo objects of a particular class, as well as any Neo object that can hold
    the object.
    Objects are searched recursively, so the objects can be nested (such as a
    list of blocks).

    Parameters
    ----------
    container : list, tuple, iterable, dict, neo.Container
                The container for the Neo objects.
    class_name : str
                The name of the class, with proper capitalization
                (i.e., 'SpikeTrain', not 'Spiketrain' or 'spiketrain').

    Returns
    -------
    list
        A list of unique Neo objects.

    Raises
    ------
    ValueError
        If can not handle containers of the type passed in `container`.

    """
    if container.__class__.__name__ == class_name:
        return [container]
    classholder = class_name.lower() + 's'
    if hasattr(container, classholder):
        vals = getattr(container, classholder)
    elif hasattr(container, 'list_children_by_class'):
        vals = container.list_children_by_class(class_name)
    elif hasattr(container, 'values') and not hasattr(container, 'ndim'):
        vals = container.values()
    elif hasattr(container, '__iter__') and not hasattr(container, 'ndim'):
        vals = container
    else:
        raise ValueError('Cannot handle object of type %s' % type(container))
    res = list(chain.from_iterable(_get_all_objs(obj, class_name)
                                   for obj in vals))
    return unique_objs(res)


def get_all_spiketrains(container):
    """
    Get all `neo.Spiketrain` objects from a container.

    The objects can be any list, dict, or other iterable or mapping containing
    spiketrains, as well as any Neo object that can hold spiketrains:
    `neo.Block`, `neo.ChannelIndex`, `neo.Unit`, and `neo.Segment`.

    Containers are searched recursively, so the objects can be nested
    (such as a list of blocks).

    Parameters
    ----------
    container : list, tuple, iterable, dict, neo.Block, neo.Segment, neo.Unit,
        neo.ChannelIndex
        The container for the spiketrains.

    Returns
    -------
    list
        A `neo.SpikeTrainList` object of the unique `neo.SpikeTrain` objects in `container`.

    """
    return SpikeTrainList(_get_all_objs(container, 'SpikeTrain'))


def get_all_events(container):
    """
    Get all `neo.Event` objects from a container.

    The objects can be any list, dict, or other iterable or mapping containing
    events, as well as any neo object that can hold events:
    `neo.Block` and `neo.Segment`.

    Containers are searched recursively, so the objects can be nested
    (such as a list of blocks).

    Parameters
    ----------
    container : list, tuple, iterable, dict, neo.Block, neo.Segment
                The container for the events.

    Returns
    -------
    list
        A list of the unique `neo.Event` objects in `container`.

    """
    return _get_all_objs(container, 'Event')


def get_all_epochs(container):
    """
    Get all `neo.Epoch` objects from a container.

    The objects can be any list, dict, or other iterable or mapping containing
    epochs, as well as any neo object that can hold epochs:
    `neo.Block` and `neo.Segment`.

    Containers are searched recursively, so the objects can be nested
    (such as a list of blocks).

    Parameters
    ----------
    container : list, tuple, iterable, dict, neo.Block, neo.Segment
                The container for the epochs.

    Returns
    -------
    list
        A list of the unique `neo.Epoch` objects in `container`.

    """
    return _get_all_objs(container, 'Epoch')



import unittest

import numpy as np
import quantities as pq
from numpy.testing import assert_array_almost_equal

from elephant.parallel import SingleProcess, ProcessPoolExecutor
from elephant.spike_train_generation import homogeneous_poisson_process
from elephant.statistics import mean_firing_rate


class TestParallel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.executors_cls = [SingleProcess, ProcessPoolExecutor]
        try:
            from elephant.parallel.mpi import MPIPoolExecutor, MPICommExecutor
            cls.executors_cls.extend([MPIPoolExecutor, MPICommExecutor])
        except ImportError:
            # mpi4py is not installed
            pass

        np.random.seed(28)
        n_spiketrains = 10
        cls.spiketrains = tuple(
            homogeneous_poisson_process(
                rate=10 * pq.Hz, t_stop=10 * pq.s, as_array=True)
            for _ in range(n_spiketrains)
        )
        cls.mean_fr = tuple(map(mean_firing_rate, cls.spiketrains))

    def test_mean_firing_rate(self):
        for executor_cls in self.executors_cls:
            with self.subTest(executor_cls=executor_cls):
                executor = executor_cls()
                mean_fr = executor.execute(handler=mean_firing_rate,
                                           args_iterate=self.spiketrains)
                assert_array_almost_equal(mean_fr, self.mean_fr)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Unit tests for the conversion module.

:copyright: Copyright 2014-2023 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest

import neo
from neo.core.spiketrainlist import SpikeTrainList
import numpy as np
import quantities as pq
from numpy.testing import (assert_array_almost_equal, assert_array_equal)

import elephant.conversion as cv
from elephant.utils import get_common_start_stop_times
from elephant.spike_train_generation import homogeneous_poisson_process


def get_nearest(times, time):
    return (np.abs(times - time)).argmin()


class binarize_TestCase(unittest.TestCase):
    def setUp(self):
        self.test_array_1d = np.array([1.23, 0.3, 0.87, 0.56])

    def test_binarize_with_spiketrain_exact(self):
        st = neo.SpikeTrain(self.test_array_1d, units='ms',
                            t_stop=10.0, sampling_rate=100)
        times = np.arange(0, 10. + .01, .01)
        target = np.zeros_like(times).astype('bool')
        for time in self.test_array_1d:
            target[get_nearest(times, time)] = True
        times = pq.Quantity(times, units='ms')

        res, tres = cv.binarize(st, return_times=True)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_spiketrain_exact_set_ends(self):
        st = neo.SpikeTrain(self.test_array_1d, units='ms',
                            t_stop=10.0, sampling_rate=100)
        times = np.arange(5., 10. + .01, .01)
        target = np.zeros_like(times).astype('bool')
        times = pq.Quantity(times, units='ms')

        res, tres = cv.binarize(st, return_times=True, t_start=5., t_stop=10.)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_spiketrain_round(self):
        st = neo.SpikeTrain(self.test_array_1d, units='ms',
                            t_stop=10.0, sampling_rate=10.0)
        times = np.arange(0, 10. + .1, .1)
        target = np.zeros_like(times).astype('bool')
        for time in np.round(self.test_array_1d, 1):
            target[get_nearest(times, time)] = True
        times = pq.Quantity(times, units='ms')

        res, tres = cv.binarize(st, return_times=True)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_quantities_exact(self):
        st = pq.Quantity(self.test_array_1d, units='ms')
        times = np.arange(0, 1.23 + .01, .01)
        target = np.zeros_like(times).astype('bool')
        for time in self.test_array_1d:
            target[get_nearest(times, time)] = True
        times = pq.Quantity(times, units='ms')

        res, tres = cv.binarize(st, return_times=True,
                                sampling_rate=100. * pq.kHz)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_quantities_exact_set_ends(self):
        st = pq.Quantity(self.test_array_1d, units='ms')
        times = np.arange(0, 10. + .01, .01)
        target = np.zeros_like(times).astype('bool')
        for time in self.test_array_1d:
            target[get_nearest(times, time)] = True
        times = pq.Quantity(times, units='ms')

        res, tres = cv.binarize(st, return_times=True, t_stop=10.,
                                sampling_rate=100. * pq.kHz)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_quantities_round_set_ends(self):
        st = pq.Quantity(self.test_array_1d, units='ms')
        times = np.arange(5., 10. + .1, .1)
        target = np.zeros_like(times).astype('bool')
        times = pq.Quantity(times, units='ms')

        res, tres = cv.binarize(st, return_times=True, t_start=5., t_stop=10.,
                                sampling_rate=10. * pq.kHz)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_plain_array_exact(self):
        st = self.test_array_1d
        times = np.arange(0, 1.23 + .01, .01)
        target = np.zeros_like(times).astype('bool')
        for time in self.test_array_1d:
            target[get_nearest(times, time)] = True

        res, tres = cv.binarize(st, return_times=True, sampling_rate=100)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_with_plain_array_exact_set_ends(self):
        st = self.test_array_1d
        times = np.arange(0, 10. + .01, .01)
        target = np.zeros_like(times).astype('bool')
        for time in self.test_array_1d:
            target[get_nearest(times, time)] = True

        res, tres = cv.binarize(st, return_times=True, t_stop=10.,
                                sampling_rate=100.)
        assert_array_almost_equal(res, target, decimal=9)
        assert_array_almost_equal(tres, times, decimal=9)

    def test_binarize_no_time(self):
        st = self.test_array_1d
        times = np.arange(0, 1.23 + .01, .01)
        target = np.zeros_like(times).astype('bool')
        for time in self.test_array_1d:
            target[get_nearest(times, time)] = True

        res0, tres = cv.binarize(st, return_times=True, sampling_rate=100)
        res1 = cv.binarize(st, return_times=False, sampling_rate=100)
        res2 = cv.binarize(st, sampling_rate=100)
        assert_array_almost_equal(res0, res1, decimal=9)
        assert_array_almost_equal(res0, res2, decimal=9)

    def test_binariz_rate_with_plain_array_and_units_typeerror(self):
        st = self.test_array_1d
        self.assertRaises(TypeError, cv.binarize, st,
                          t_start=pq.Quantity(0, 'ms'),
                          sampling_rate=10.)
        self.assertRaises(TypeError, cv.binarize, st,
                          t_stop=pq.Quantity(10, 'ms'),
                          sampling_rate=10.)
        self.assertRaises(TypeError, cv.binarize, st,
                          t_start=pq.Quantity(0, 'ms'),
                          t_stop=pq.Quantity(10, 'ms'),
                          sampling_rate=10.)
        self.assertRaises(TypeError, cv.binarize, st,
                          t_start=pq.Quantity(0, 'ms'),
                          t_stop=10.,
                          sampling_rate=10.)
        self.assertRaises(TypeError, cv.binarize, st,
                          t_start=0.,
                          t_stop=pq.Quantity(10, 'ms'),
                          sampling_rate=10.)
        self.assertRaises(TypeError, cv.binarize, st,
                          sampling_rate=10. * pq.Hz)

    def test_binariz_without_sampling_rate_valueerror(self):
        st0 = self.test_array_1d
        st1 = pq.Quantity(st0, 'ms')
        self.assertRaises(ValueError, cv.binarize, st0)
        self.assertRaises(ValueError, cv.binarize, st0,
                          t_start=0)
        self.assertRaises(ValueError, cv.binarize, st0,
                          t_stop=10)
        self.assertRaises(ValueError, cv.binarize, st0,
                          t_start=0, t_stop=10)
        self.assertRaises(ValueError, cv.binarize, st1,
                          t_start=pq.Quantity(0, 'ms'), t_stop=10.)
        self.assertRaises(ValueError, cv.binarize, st1,
                          t_start=0., t_stop=pq.Quantity(10, 'ms'))
        self.assertRaises(ValueError, cv.binarize, st1)

    def test_bin_edges_empty_binned_spiketrain(self):
        st = neo.SpikeTrain(times=np.array([2.5]) * pq.s, t_start=0 * pq.s,
                            t_stop=3 * pq.s)
        with self.assertWarns(UserWarning):
            bst = cv.BinnedSpikeTrain(st, bin_size=2 * pq.s, t_start=0 * pq.s,
                                      t_stop=3 * pq.s)
        assert_array_equal(bst.bin_edges, [0., 2.] * pq.s)
        assert_array_equal(bst.spike_indices, [[]])  # no binned spikes
        self.assertEqual(bst.get_num_of_spikes(), 0)

    def test_regression_431(self):
        """
        Addresses issue 431
        This unittest addresses an issue where a SpikeTrainList obejct was not
        correctly handled by the constructor
        """
        st1 = neo.SpikeTrain(
            times=np.array([1, 2, 3]) * pq.ms,
            t_start=0 * pq.ms, t_stop=10 * pq.ms)
        st2 = neo.SpikeTrain(
            times=np.array([4, 5, 6]) * pq.ms,
            t_start=0 * pq.ms, t_stop=10 * pq.ms)
        real_list = [st1, st2]
        spiketrainlist = SpikeTrainList([st1, st2])

        real_list_binary = cv.BinnedSpikeTrain(real_list, bin_size=1*pq.ms)
        spiketrainlist_binary = cv.BinnedSpikeTrain(
            spiketrainlist, bin_size=1 * pq.ms)

        assert_array_equal(
            real_list_binary.to_array(), spiketrainlist_binary.to_array())


class BinnedSpikeTrainTestCase(unittest.TestCase):
    def setUp(self):
        self.spiketrain_a = neo.SpikeTrain(
            [0.5, 0.7, 1.2, 3.1, 4.3, 5.5, 6.7] * pq.s, t_stop=10.0 * pq.s)
        self.spiketrain_b = neo.SpikeTrain(
            [0.1, 0.7, 1.2, 2.2, 4.3, 5.5, 8.0] * pq.s, t_stop=10.0 * pq.s)
        self.bin_size = 1 * pq.s
        self.tolerance = 1e-8

    def test_binarize(self):
        spiketrains = [self.spiketrain_a, self.spiketrain_b,
                       self.spiketrain_a, self.spiketrain_b]
        for sparse_format in ("csr", "csc"):
            bst = cv.BinnedSpikeTrain(spiketrains=spiketrains,
                                      bin_size=self.bin_size,
                                      sparse_format=sparse_format)
            bst_bin = bst.binarize(copy=True)
            bst_copy = bst.copy()
            assert_array_equal(bst_bin.to_array(), bst.to_bool_array())
            bst_copy.sparse_matrix.data[:] = 1
            self.assertEqual(bst_bin, bst_copy)

    def test_slice(self):
        spiketrains = [self.spiketrain_a, self.spiketrain_b,
                       self.spiketrain_a, self.spiketrain_b]
        bst = cv.BinnedSpikeTrain(spiketrains=spiketrains,
                                  bin_size=self.bin_size)
        self.assertEqual(bst[:, :], bst)
        self.assertEqual(bst[1:], cv.BinnedSpikeTrain(spiketrains[1:],
                                                      bin_size=self.bin_size))
        self.assertEqual(bst[:, :4], bst.time_slice(t_stop=4 * pq.s))
        self.assertEqual(bst[:, 1:-1], cv.BinnedSpikeTrain(
            spiketrains, bin_size=self.bin_size,
            t_start=1 * pq.s, t_stop=9 * pq.s
        ))
        self.assertEqual(bst[0, 0], cv.BinnedSpikeTrain(
            neo.SpikeTrain([0.5, 0.7], t_stop=1, units='s'),
            bin_size=self.bin_size
        ))

        # 2-seconds stride: leave [0..1, 2..3, 4..5, 6..7] interval
        self.assertEqual(bst[0, ::2], cv.BinnedSpikeTrain(
            neo.SpikeTrain([0.5, 0.7, 4.3, 6.7], t_stop=10, units='s'),
            bin_size=2 * self.bin_size
        ))

        bst_copy = bst.copy()
        bst_copy[:] = 1
        assert_array_equal(bst_copy.sparse_matrix.todense(), 1)

    def test_time_slice(self):
        spiketrains = [self.spiketrain_a, self.spiketrain_b]
        bst = cv.BinnedSpikeTrain(spiketrains=spiketrains,
                                  bin_size=self.bin_size)
        bst_equal = bst.time_slice(t_start=bst.t_start - 5 * pq.s,
                                   t_stop=bst.t_stop + 5 * pq.s)
        self.assertEqual(bst_equal, bst)
        bst_same = bst.time_slice(t_start=None, t_stop=None)
        self.assertIs(bst_same, bst)
        bst_copy = bst.time_slice(t_start=None, t_stop=None, copy=True)
        self.assertIsNot(bst_copy, bst)
        self.assertEqual(bst_copy, bst)
        bst_empty = bst.time_slice(t_start=0.2 * pq.s, t_stop=0.3 * pq.s)
        self.assertEqual(bst_empty.n_bins, 0)
        t_range = np.arange(0, 10, self.bin_size.item()) * pq.s
        for i, t_start in enumerate(t_range[:-1]):
            for t_stop in t_range[i + 1:]:
                bst_ij = bst.time_slice(t_start=t_start, t_stop=t_stop)
                bst_ij2 = bst_ij.time_slice(t_start=t_start, t_stop=t_stop)
                self.assertEqual(bst_ij2, bst_ij)
                self.assertEqual(bst_ij2.tolerance, bst.tolerance)
                sts = [st.time_slice(t_start=t_start, t_stop=t_stop)
                       for st in spiketrains]
                bst_ref = cv.BinnedSpikeTrain(sts, bin_size=self.bin_size)
                self.assertEqual(bst_ij, bst_ref)

        # invalid input: not a quantity
        self.assertRaises(TypeError, bst.time_slice, t_start=2)

    def test_to_spike_trains(self):
        np.random.seed(1)
        spiketrains = [homogeneous_poisson_process(rate=10 * pq.Hz,
                                                   t_start=-1 * pq.s,
                                                   t_stop=10 * pq.s)]
        for sparse_format in ("csr", "csc"):
            bst1 = cv.BinnedSpikeTrain(
                spiketrains=[self.spiketrain_a, self.spiketrain_b],
                bin_size=self.bin_size, sparse_format=sparse_format
            )
            bst2 = cv.BinnedSpikeTrain(spiketrains=spiketrains,
                                       bin_size=300 * pq.ms,
                                       sparse_format=sparse_format)
            for bst in (bst1, bst2):
                for spikes in ("random", "left", "center"):
                    spiketrains_gen = bst.to_spike_trains(spikes=spikes,
                                                          annotate_bins=True)
                    for st, indices in zip(spiketrains_gen, bst.spike_indices):
                        # check sorted
                        self.assertTrue((np.diff(st.magnitude) > 0).all())
                        assert_array_equal(st.array_annotations['bins'],
                                           indices)
                        self.assertEqual(st.annotations['bin_size'],
                                         bst.bin_size)
                        self.assertEqual(st.t_start, bst.t_start)
                        self.assertEqual(st.t_stop, bst.t_stop)
                    bst_same = cv.BinnedSpikeTrain(spiketrains_gen,
                                                   bin_size=bst.bin_size,
                                                   sparse_format=sparse_format)
                    self.assertEqual(bst_same, bst)

                # invalid mode
                self.assertRaises(ValueError, bst.to_spike_trains,
                                  spikes='right')

    def test_get_num_of_spikes(self):
        spiketrains = [self.spiketrain_a, self.spiketrain_b]
        for spiketrain in spiketrains:
            binned = cv.BinnedSpikeTrain(spiketrain, n_bins=10,
                                         bin_size=1 * pq.s, t_start=0 * pq.s)
            self.assertEqual(binned.get_num_of_spikes(),
                             len(binned.spike_indices[0]))
        for sparse_format in ("csr", "csc"):
            binned_matrix = cv.BinnedSpikeTrain(spiketrains, n_bins=10,
                                                bin_size=1 * pq.s,
                                                sparse_format=sparse_format)
            n_spikes_per_row = binned_matrix.get_num_of_spikes(axis=1)
            n_spikes_per_row_from_indices = list(
                map(len, binned_matrix.spike_indices))
            assert_array_equal(n_spikes_per_row, n_spikes_per_row_from_indices)
            self.assertEqual(binned_matrix.get_num_of_spikes(),
                             sum(n_spikes_per_row_from_indices))

    def test_binned_spiketrain_sparse(self):
        a = neo.SpikeTrain([1.7, 1.8, 4.3] * pq.s, t_stop=10.0 * pq.s)
        b = neo.SpikeTrain([1.7, 1.8, 4.3] * pq.s, t_stop=10.0 * pq.s)
        bin_size = 1 * pq.s
        nbins = 10
        x = cv.BinnedSpikeTrain([a, b], n_bins=nbins, bin_size=bin_size,
                                t_start=0 * pq.s)
        x_sparse = [2, 1, 2, 1]
        assert_array_equal(x.sparse_matrix.data, x_sparse)
        assert_array_equal(x.spike_indices, [[1, 1, 4], [1, 1, 4]])

    def test_binned_spiketrain_shape(self):
        a = self.spiketrain_a
        x = cv.BinnedSpikeTrain(a, n_bins=10,
                                bin_size=self.bin_size,
                                t_start=0 * pq.s)
        x_bool = cv.BinnedSpikeTrain(a, n_bins=10, bin_size=self.bin_size,
                                     t_start=0 * pq.s)
        self.assertEqual(x.to_array().shape, (1, 10))
        self.assertEqual(x_bool.to_bool_array().shape, (1, 10))

    # shape of the matrix for a list of spike trains
    def test_binned_spiketrain_shape_list(self):
        a = self.spiketrain_a
        b = self.spiketrain_b
        c = [a, b]
        nbins = 5
        x = cv.BinnedSpikeTrain(c, n_bins=nbins, t_start=0 * pq.s,
                                t_stop=10.0 * pq.s)
        x_bool = cv.BinnedSpikeTrain(c, n_bins=nbins, t_start=0 * pq.s,
                                     t_stop=10.0 * pq.s)
        self.assertEqual(x.to_array().shape, (2, 5))
        self.assertEqual(x_bool.to_bool_array().shape, (2, 5))

    def test_binned_spiketrain_neg_times(self):
        a = neo.SpikeTrain(
            [-6.5, 0.5, 0.7, 1.2, 3.1, 4.3, 5.5, 6.7] * pq.s,
            t_start=-6.5 * pq.s, t_stop=10.0 * pq.s)
        bin_size = self.bin_size
        nbins = 16
        x = cv.BinnedSpikeTrain(a, n_bins=nbins, bin_size=bin_size,
                                t_start=-6.5 * pq.s)
        y = [[1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0]]
        assert_array_equal(x.to_bool_array(), y)

    def test_binned_spiketrain_neg_times_list(self):
        a = neo.SpikeTrain(
            [-6.5, 0.5, 0.7, 1.2, 3.1, 4.3, 5.5, 6.7] * pq.s,
            t_start=-7 * pq.s, t_stop=7 * pq.s)
        b = neo.SpikeTrain(
            [-0.1, -0.7, 1.2, 2.2, 4.3, 5.5, 8.0] * pq.s,
            t_start=-1 * pq.s, t_stop=8 * pq.s)
        spiketrains = [a, b]

        # not the same t_start and t_stop
        self.assertRaises(ValueError, cv.BinnedSpikeTrain,
                          spiketrains=spiketrains,
                          bin_size=self.bin_size)
        t_start, t_stop = get_common_start_stop_times(spiketrains)
        self.assertEqual(t_start, -1 * pq.s)
        self.assertEqual(t_stop, 7 * pq.s)
        x_bool = cv.BinnedSpikeTrain(spiketrains, bin_size=self.bin_size,
                                     t_start=t_start, t_stop=t_stop)
        y_bool = [[0, 1, 1, 0, 1, 1, 1, 1],
                  [1, 0, 1, 1, 0, 1, 1, 0]]

        assert_array_equal(x_bool.to_bool_array(), y_bool)

    # checking spike_indices(f) and matrix(m) for 1 spiketrain
    def test_binned_spiketrain_indices(self):
        a = self.spiketrain_a
        bin_size = self.bin_size
        nbins = 10
        x = cv.BinnedSpikeTrain(a, n_bins=nbins, bin_size=bin_size,
                                t_start=0 * pq.s)
        x_bool = cv.BinnedSpikeTrain(a, n_bins=nbins, bin_size=bin_size,
                                     t_start=0 * pq.s)
        y_matrix = [[2., 1., 0., 1., 1., 1., 1., 0., 0., 0.]]
        y_bool_matrix = [[1., 1., 0., 1., 1., 1., 1., 0., 0., 0.]]
        assert_array_equal(x.to_array(), y_matrix)
        assert_array_equal(x_bool.to_bool_array(), y_bool_matrix)
        s = x_bool.to_sparse_bool_array()[
            x_bool.to_sparse_bool_array().nonzero()]
        assert_array_equal(s, [[True] * 6])

    def test_binned_spiketrain_list(self):
        a = self.spiketrain_a
        b = self.spiketrain_b

        bin_size = self.bin_size
        nbins = 10
        c = [a, b]
        x = cv.BinnedSpikeTrain(c, n_bins=nbins, bin_size=bin_size,
                                t_start=0 * pq.s)
        x_bool = cv.BinnedSpikeTrain(c, n_bins=nbins, bin_size=bin_size,
                                     t_start=0 * pq.s)
        y_matrix = [[2, 1, 0, 1, 1, 1, 1, 0, 0, 0],
                    [2, 1, 1, 0, 1, 1, 0, 0, 1, 0]]
        y_matrix_bool = [[1, 1, 0, 1, 1, 1, 1, 0, 0, 0],
                         [1, 1, 1, 0, 1, 1, 0, 0, 1, 0]]
        assert_array_equal(x.to_array(), y_matrix)
        assert_array_equal(x_bool.to_bool_array(), y_matrix_bool)

    # t_stop is None
    def test_binned_spiketrain_list_t_stop(self):
        a = self.spiketrain_a
        b = self.spiketrain_b
        c = [a, b]
        bin_size = self.bin_size
        nbins = 10
        x = cv.BinnedSpikeTrain(c, n_bins=nbins, bin_size=bin_size,
                                t_start=0 * pq.s,
                                t_stop=None)
        x_bool = cv.BinnedSpikeTrain(c, n_bins=nbins, bin_size=bin_size,
                                     t_start=0 * pq.s)
        self.assertEqual(x.t_stop, 10 * pq.s)
        self.assertEqual(x_bool.t_stop, 10 * pq.s)

    # Test number of bins
    def test_binned_spiketrain_list_numbins(self):
        a = self.spiketrain_a
        b = self.spiketrain_b
        c = [a, b]
        bin_size = 1 * pq.s
        x = cv.BinnedSpikeTrain(c, bin_size=bin_size, t_start=0 * pq.s,
                                t_stop=10. * pq.s)
        x_bool = cv.BinnedSpikeTrain(c, bin_size=bin_size, t_start=0 * pq.s,
                                     t_stop=10. * pq.s)
        self.assertEqual(x.n_bins, 10)
        self.assertEqual(x_bool.n_bins, 10)

    def test_binned_spiketrain_matrix(self):
        # Init
        a = self.spiketrain_a
        b = self.spiketrain_b
        x_bool_a = cv.BinnedSpikeTrain(a, bin_size=pq.s, t_start=0 * pq.s,
                                       t_stop=10. * pq.s)
        x_bool_b = cv.BinnedSpikeTrain(b, bin_size=pq.s, t_start=0 * pq.s,
                                       t_stop=10. * pq.s)

        # Assumed results
        y_matrix_a = [[2, 1, 0, 1, 1, 1, 1, 0, 0, 0]]
        y_matrix_bool_a = [[1, 1, 0, 1, 1, 1, 1, 0, 0, 0]]
        y_matrix_bool_b = [[1, 1, 1, 0, 1, 1, 0, 0, 1, 0]]

        # Asserts
        assert_array_equal(x_bool_a.to_bool_array(), y_matrix_bool_a)
        assert_array_equal(x_bool_b.to_bool_array(), y_matrix_bool_b)
        assert_array_equal(x_bool_a.to_array(), y_matrix_a)

    # Test if t_start is calculated correctly
    def test_binned_spiketrain_parameter_calc_tstart(self):
        x = cv.BinnedSpikeTrain(self.spiketrain_a, bin_size=1 * pq.s,
                                n_bins=10, t_stop=10. * pq.s)
        self.assertEqual(x.t_start, 0. * pq.s)
        self.assertEqual(x.t_stop, 10. * pq.s)
        self.assertEqual(x.bin_size, 1 * pq.s)
        self.assertEqual(x.n_bins, 10)

    # Test if error raises when type of n_bins is not an integer
    def test_binned_spiketrain_n_bins_not_int(self):
        a = self.spiketrain_a
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a, bin_size=pq.s,
                          n_bins=1.4, t_start=0 * pq.s,
                          t_stop=10. * pq.s)

    def test_to_array(self):
        x = cv.BinnedSpikeTrain(self.spiketrain_a, bin_size=1 * pq.s,
                                n_bins=10, t_stop=10. * pq.s)
        arr_float = x.to_array(dtype=float)
        assert_array_equal(arr_float, x.to_array().astype(float))

    # Test if error is raised when providing insufficient number of
    # parameters
    def test_binned_spiketrain_insufficient_arguments(self):
        a = self.spiketrain_a
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a)
        self.assertRaises(
            ValueError,
            cv.BinnedSpikeTrain,
            a,
            bin_size=1 * pq.s,
            t_start=0 * pq.s,
            t_stop=0 * pq.s)

    def test_different_input_types(self):
        a = self.spiketrain_a
        q = [1, 2, 3] * pq.s
        self.assertRaises(ValueError, cv.BinnedSpikeTrain,
                          spiketrains=[a, q], bin_size=pq.s)

    def test_get_start_stop(self):
        a = self.spiketrain_a
        b = neo.SpikeTrain(
            [-0.1, -0.7, 1.2, 2.2, 4.3, 5.5, 8.0] * pq.s,
            t_start=-1 * pq.s, t_stop=8 * pq.s)
        start, stop = get_common_start_stop_times(a)
        self.assertEqual(start, a.t_start)
        self.assertEqual(stop, a.t_stop)
        start, stop = get_common_start_stop_times([a, b])
        self.assertEqual(start, a.t_start)
        self.assertEqual(stop, b.t_stop)

    def test_consistency_errors(self):
        a = self.spiketrain_a
        b = neo.SpikeTrain([-2, -1] * pq.s, t_start=-2 * pq.s,
                           t_stop=-1 * pq.s)
        self.assertRaises(TypeError, cv.BinnedSpikeTrain, [a, b], t_start=5,
                          t_stop=0, bin_size=pq.s, n_bins=10)

        b = neo.SpikeTrain([-7, -8, -9] * pq.s, t_start=-9 * pq.s,
                           t_stop=-7 * pq.s)
        self.assertRaises(TypeError, cv.BinnedSpikeTrain, b, t_start=None,
                          t_stop=10, bin_size=pq.s, n_bins=10)
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a, t_start=0 * pq.s,
                          t_stop=10 * pq.s, bin_size=3 * pq.s, n_bins=10)

        b = neo.SpikeTrain([-4, -2, 0, 1] * pq.s, t_start=-4 * pq.s,
                           t_stop=1 * pq.s)
        self.assertRaises(
            TypeError,
            cv.BinnedSpikeTrain,
            b,
            bin_size=-2 * pq.s,
            t_start=-4 * pq.s,
            t_stop=0 * pq.s)

    # Test edges
    def test_binned_spiketrain_bin_edges(self):
        a = self.spiketrain_a
        x = cv.BinnedSpikeTrain(a, bin_size=1 * pq.s, n_bins=10,
                                t_stop=10. * pq.s)
        # Test all edges
        assert_array_equal(x.bin_edges, [float(i) for i in range(11)])

        # Test center edges
        assert_array_equal(x.bin_centers, np.arange(0, 10) + 0.5)

    # Test for different units but same times
    def test_binned_spiketrain_different_units(self):
        a = self.spiketrain_a
        b = a.rescale(pq.ms)
        bin_size = 1 * pq.s
        xa = cv.BinnedSpikeTrain(a, bin_size=bin_size)
        xb = cv.BinnedSpikeTrain(b, bin_size=bin_size.rescale(pq.ms))
        assert_array_equal(xa.to_array(), xb.to_array())
        assert_array_equal(xa.to_bool_array(), xb.to_bool_array())
        assert_array_equal(xa.sparse_matrix.data,
                           xb.sparse_matrix.data)
        assert_array_equal(xa.bin_edges, xb.bin_edges)

    def test_binary_to_binned_matrix(self):
        a = [[1, 0, 0, 0], [0, 1, 1, 0]]
        x = cv.BinnedSpikeTrain(a, t_start=0 * pq.s, t_stop=5 * pq.s)
        # Check for correctness with different init params
        assert_array_equal(x.to_array(), a)
        assert_array_equal(x.to_bool_array(), a)
        self.assertEqual(x.n_bins, 4)
        self.assertEqual(x.bin_size, 1.25 * pq.s)

        x = cv.BinnedSpikeTrain(a, t_start=1 * pq.s, bin_size=2 * pq.s)
        assert_array_equal(x.to_array(), a)
        assert_array_equal(x.to_bool_array(), a)
        self.assertEqual(x.t_stop, 9 * pq.s)

        x = cv.BinnedSpikeTrain(a, t_stop=9 * pq.s, bin_size=2 * pq.s)
        self.assertEqual(x.t_start, 1 * pq.s)

        # Raise error
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a,
                          t_start=5 * pq.s, t_stop=0 * pq.s, bin_size=pq.s,
                          n_bins=10)
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a, t_start=0 * pq.s,
                          t_stop=10 * pq.s, bin_size=3 * pq.s, n_bins=10)
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a,
                          bin_size=-2 * pq.s, t_start=-4 * pq.s,
                          t_stop=0 * pq.s)

        # Check binary property
        self.assertTrue(x.is_binary)

    def test_binned_to_binned(self):
        a = self.spiketrain_a
        x = cv.BinnedSpikeTrain(a, bin_size=1 * pq.s).to_array()
        y = cv.BinnedSpikeTrain(x, bin_size=1 * pq.s, t_start=0 * pq.s)
        assert_array_equal(y.to_array(), x)

        # test with a list
        x = cv.BinnedSpikeTrain([[0, 1, 2, 3]], bin_size=1 * pq.s,
                                t_stop=3 * pq.s).to_array()
        y = cv.BinnedSpikeTrain(x, bin_size=1 * pq.s, t_start=0 * pq.s)
        assert_array_equal(y.to_array(), x)

        # test with a numpy array
        a = np.array([[0, 1, 2, 3], [1, 2, 2.5, 3]])
        x = cv.BinnedSpikeTrain(a, bin_size=1 * pq.s,
                                t_stop=3 * pq.s).to_array()
        y = cv.BinnedSpikeTrain(x, bin_size=1 * pq.s, t_start=0 * pq.s)
        assert_array_equal(y.to_array(), x)

        # Check binary property
        self.assertFalse(y.is_binary)

        # Raise Errors
        # give a strangely shaped matrix as input (not MxN)
        a = np.array([[0, 1, 2, 3], [1, 2, 3]], dtype=object)
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a, t_start=0 * pq.s,
                          bin_size=1 * pq.s)
        # Give no t_start or t_stop
        a = np.array([[0, 1, 2, 3], [1, 2, 3, 4]])
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a,
                          bin_size=1 * pq.s)
        # Input format not supported
        a = np.array(([0, 1, 2], [0, 1, 2, 3, 4]), dtype=object)
        self.assertRaises(ValueError, cv.BinnedSpikeTrain, a,
                          bin_size=1 * pq.s)

    def test_binnend_spiketrain_different_input_units(self):
        train = neo.SpikeTrain(times=np.array([1.001, 1.002, 1.005]) * pq.s,
                               t_start=1 * pq.s, t_stop=1.01 * pq.s)
        bst = cv.BinnedSpikeTrain(train,
                                  t_start=1 * pq.s, t_stop=1.01 * pq.s,
                                  bin_size=1 * pq.ms)
        self.assertEqual(bst.units, pq.s)
        target_edges = np.array([1000, 1001, 1002, 1003, 1004, 1005, 1006,
                                 1007, 1008, 1009, 1010], dtype=float
                                ) * pq.ms
        target_centers = np.array(
            [1000.5, 1001.5, 1002.5, 1003.5, 1004.5, 1005.5, 1006.5, 1007.5,
             1008.5, 1009.5], dtype=float) * pq.ms
        assert_array_almost_equal(bst.bin_edges, target_edges)
        assert_array_almost_equal(bst.bin_centers, target_centers)

        bst = cv.BinnedSpikeTrain(train,
                                  t_start=1 * pq.s, t_stop=1010 * pq.ms,
                                  bin_size=1 * pq.ms)
        self.assertEqual(bst.units, pq.s)
        assert_array_almost_equal(bst.bin_edges, target_edges)
        assert_array_almost_equal(bst.bin_centers, target_centers)

    def test_rescale(self):
        train = neo.SpikeTrain(times=np.array([1.001, 1.002, 1.005]) * pq.s,
                               t_start=1 * pq.s, t_stop=1.01 * pq.s)
        bst = cv.BinnedSpikeTrain(train, t_start=1 * pq.s,
                                  t_stop=1.01 * pq.s,
                                  bin_size=1 * pq.ms)
        self.assertEqual(bst.units, pq.s)
        self.assertEqual(bst._t_start, 1)  # 1 s
        self.assertEqual(bst._t_stop, 1.01)  # 1.01 s
        self.assertEqual(bst._bin_size, 0.001)  # 0.001 s

        bst.rescale(units='ms')
        self.assertEqual(bst.units, pq.ms)
        self.assertEqual(bst._t_start, 1000)  # 1 s
        self.assertEqual(bst._t_stop, 1010)  # 1.01 s
        self.assertEqual(bst._bin_size, 1)  # 0.001 s

    def test_repr(self):
        train = neo.SpikeTrain(times=np.array([1.001, 1.002, 1.005]) * pq.s,
                               t_start=1 * pq.s, t_stop=1.01 * pq.s)
        bst = cv.BinnedSpikeTrain(train, t_start=1 * pq.s,
                                  t_stop=1.01 * pq.s,
                                  bin_size=1 * pq.ms)
        self.assertEqual(repr(bst), "BinnedSpikeTrain(t_start=1.0 s, "
                                    "t_stop=1.01 s, bin_size=0.001 s; "
                                    "shape=(1, 10), format=csr_matrix)")

    def test_binned_sparsity(self):
        train = neo.SpikeTrain(np.arange(10), t_stop=10 * pq.s, units=pq.s)
        bst = cv.BinnedSpikeTrain(train, n_bins=100)
        self.assertAlmostEqual(bst.sparsity, 0.1)

    # Test fix for rounding errors
    def test_binned_spiketrain_rounding(self):
        train = neo.SpikeTrain(times=np.arange(120000) / 30000. * pq.s,
                               t_start=0 * pq.s, t_stop=4 * pq.s)
        bst = cv.BinnedSpikeTrain(train,
                                  t_start=0 * pq.s, t_stop=4 * pq.s,
                                  bin_size=1. / 30000. * pq.s)
        assert_array_equal(bst.to_array().nonzero()[1],
                           np.arange(120000))


class DiscretiseSpiketrainsTestCase(unittest.TestCase):
    def setUp(self):
        times = (np.arange(10) + np.random.uniform(size=10)) * pq.ms
        self.spiketrains = [neo.SpikeTrain(times, t_stop=10*pq.ms)] * 5

    def test_list_of_spiketrains(self):
        discretised_spiketrains = cv.discretise_spiketimes(self.spiketrains,
                                                           1 / pq.ms)
        for idx in range(len(self.spiketrains)):
            np.testing.assert_array_equal(discretised_spiketrains[idx].times,
                                          np.arange(10) * pq.ms)

    def test_single_spiketrain(self):
        discretised_spiketrain = cv.discretise_spiketimes(self.spiketrains[0],
                                                          1 / pq.ms)
        np.testing.assert_array_equal(discretised_spiketrain.times,
                                      np.arange(10) * pq.ms)

    def test_preserve_t_start(self):
        spiketrain = neo.SpikeTrain([0.7, 5.1]*pq.ms,
                                    t_start=0.5*pq.ms, t_stop=10*pq.ms)
        with self.assertWarns(UserWarning):
            discretised_spiketrain = cv.discretise_spiketimes(spiketrain,
                                                              1 / pq.ms)
        np.testing.assert_array_equal(discretised_spiketrain.times,
                                      [0.5, 5] * pq.ms)

    def test_binning_consistency(self):
        discretised_spiketrains = cv.discretise_spiketimes(self.spiketrains,
                                                           1 / pq.ms)
        bsts = cv.BinnedSpikeTrain(self.spiketrains,
                                   bin_size=1 * pq.ms)
        bsts_discretised = cv.BinnedSpikeTrain(discretised_spiketrains,
                                               bin_size=1 * pq.ms)
        np.testing.assert_array_equal(bsts.to_array(),
                                      bsts_discretised.to_array())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
*\"Current Source Density analysis (CSD) is a class of methods of analysis of
extracellular electric potentials recorded at multiple sites leading to
estimates of current sources generating the measured potentials. It is usually
applied to low-frequency part of the potential (called the Local Field
Potential, LFP) and to simultaneous recordings or to recordings taken with
fixed time reference to the onset of specific stimulus (Evoked Potentials).\"*
(Definition by Prof.Daniel K. Wójcik for Encyclopedia of Computational
Neuroscience.)

CSD is also called as Source Localization or Source Imaging in the EEG circles.
Here are CSD methods for different types of electrode configurations.

- 1D -laminar probe like electrodes.
- 2D -Microelectrode Array like
- 3D -UtahArray or multiple laminar probes.

The following methods have been implemented so far

- 1D: StandardCSD, DeltaiCSD, SplineiCSD, StepiCSD, KCSD1D
- 2D: KCSD2D, MoIKCSD (Saline layer on top of slice)
- 3D: KCSD3D

Each listed method has certain advantages. The KCSD methods, for instance, can
handle broken or irregular electrode configurations electrode.

.. autosummary::
    :toctree: _toctree/current_source_density

    estimate_csd
    generate_lfp

"""

from __future__ import division, print_function, unicode_literals

import neo
import numpy as np
import quantities as pq
from scipy.integrate import simps

import elephant.current_source_density_src.utility_functions as utils
from elephant.current_source_density_src import KCSD, icsd
from elephant.utils import deprecated_alias

__all__ = [
    "estimate_csd",
    "generate_lfp"
]

utils.patch_quantities()

available_1d = ['StandardCSD', 'DeltaiCSD', 'StepiCSD', 'SplineiCSD', 'KCSD1D']
available_2d = ['KCSD2D', 'MoIKCSD']
available_3d = ['KCSD3D']

kernel_methods = ['KCSD1D', 'KCSD2D', 'KCSD3D', 'MoIKCSD']
icsd_methods = ['DeltaiCSD', 'StepiCSD', 'SplineiCSD']

py_iCSD_toolbox = ['StandardCSD'] + icsd_methods


@deprecated_alias(coords='coordinates')
def estimate_csd(lfp, coordinates='coordinates', method=None,
                 process_estimate=True, **kwargs):
    """
    Function call to compute the current source density (CSD) from
    extracellular potential recordings (local field potentials - LFP) using
    laminar electrodes or multi-contact electrodes with 2D or 3D geometries.

    Parameters
    ----------
    lfp : :class:`neo.core.AnalogSignal`
        Positions of electrodes can be added as an array annotation
    coordinates : array-like Quantity or string
        Specifies the corresponding spatial coordinates of the electrodes.
        Coordinates can be directly supplied by a NxM array-like Quantity
        with dimension of space, where M is the number of signals in 'lfp',
        and N is equal to the dimensionality of the method.
        Alternatively, if coordinates is a string, the function will fetch the
        coordinates, supplied in the same format, as annotation of 'lfp' by
        that name.
        Default: 'coordinates'
    method : string
        Pick a method corresponding to the setup, in this implementation
        For Laminar probe style (1D), use 'KCSD1D' or 'StandardCSD',
        or 'DeltaiCSD' or 'StepiCSD' or 'SplineiCSD'
        For MEA probe style (2D),  use 'KCSD2D', or 'MoIKCSD'
        For array of laminar probes (3D), use 'KCSD3D'
        Defaults to None
    process_estimate : bool
        In the py_iCSD_toolbox this corresponds to the filter_csd -
        the parameters are passed as kwargs here ie., f_type and f_order
        In the kcsd methods this corresponds to cross_validate -
        the parameters are passed as kwargs here ie., lambdas and Rs
        Defaults to True
    kwargs : parameters to each method
        The parameters corresponding to the method chosen
        See the documentation of the individual method
        Default is {} - picks the best parameters,

    Returns
    -------
    Estimated CSD
       neo.AnalogSignal object
       annotated with the spatial coordinates

    Raises
    ------
    AttributeError
        No units specified for electrode spatial coordinates
    ValueError
        Invalid function arguments, wrong method name, or
        mismatching coordinates
    TypeError
        Invalid cv_param argument passed
    """
    if not isinstance(lfp, neo.AnalogSignal):
        raise TypeError('Parameter `lfp` must be a neo.AnalogSignal object')
    if isinstance(coordinates, str):
        coordinates = lfp.annotations[coordinates]

    # Scale all coordinates to mm as common basis
    scaled_coords = []
    for coord in coordinates:
        try:
            scaled_coords.append(coord.rescale(pq.mm))
        except AttributeError:
            raise AttributeError('No units given for electrode spatial \
            coordinates')
    coordinates = scaled_coords

    if method is None:
        raise ValueError('Must specify a method of CSD implementation')
    if len(coordinates) != lfp.shape[1]:
        raise ValueError('Number of signals and coords is not same')
    for ii in coordinates:  # CHECK for Dimensionality of electrodes
        if len(ii) > 3:
            raise ValueError('Invalid number of coordinate positions')
    dim = len(coordinates[0])  # TODO : Generic co-ordinates!
    if dim == 1 and (method not in available_1d):
        raise ValueError('Invalid method, Available options are:',
                         available_1d)
    if dim == 2 and (method not in available_2d):
        raise ValueError('Invalid method, Available options are:',
                         available_2d)
    if dim == 3 and (method not in available_3d):
        raise ValueError('Invalid method, Available options are:',
                         available_3d)
    if method in kernel_methods:
        input_array = np.zeros((len(lfp), lfp[0].magnitude.shape[0]))
        for ii, jj in enumerate(lfp):
            input_array[ii, :] = jj.rescale(pq.mV).magnitude
        kernel_method = getattr(KCSD, method)  # fetch the class 'KCSD1D'
        lambdas = kwargs.pop('lambdas', None)
        Rs = kwargs.pop('Rs', None)
        k = kernel_method(np.array(coordinates), input_array.T, **kwargs)
        if process_estimate:
            k.cross_validate(lambdas, Rs)
        estm_csd = k.values()
        estm_csd = np.rollaxis(estm_csd, -1, 0)
        output = neo.AnalogSignal(estm_csd * pq.uA / pq.mm**3,
                                  t_start=lfp.t_start,
                                  sampling_rate=lfp.sampling_rate)

        if dim == 1:
            output.annotate(x_coords=k.estm_x)
        elif dim == 2:
            output.annotate(x_coords=k.estm_x, y_coords=k.estm_y)
        elif dim == 3:
            output.annotate(x_coords=k.estm_x, y_coords=k.estm_y,
                            z_coords=k.estm_z)
    elif method in py_iCSD_toolbox:

        coordinates = np.array(coordinates) * coordinates[0].units

        if method in icsd_methods:
            try:
                coordinates = coordinates.rescale(kwargs['diam'].units)
            except KeyError:  # Then why specify as a default in icsd?
                # All iCSD methods explicitly assume a source
                # diameter in contrast to the stdCSD  that
                # implicitly assume infinite source radius
                raise ValueError(f"Parameter diam must be specified for iCSD "
                                 f"methods: {', '.join(icsd_methods)}")

        if 'f_type' in kwargs:
            if (kwargs['f_type'] != 'identity') and  \
               (kwargs['f_order'] is None):
                raise ValueError(f"The order of {kwargs['f_type']} filter must"
                                 f" be specified")

        csd_method = getattr(icsd, method)  # fetch class from icsd.py file
        csd_estimator = csd_method(lfp=lfp.T.magnitude * lfp.units,
                                   coord_electrode=coordinates.flatten(),
                                   **kwargs)
        csd_pqarr = csd_estimator.get_csd()

        if process_estimate:
            csd_pqarr_filtered = csd_estimator.filter_csd(csd_pqarr)
            output = neo.AnalogSignal(csd_pqarr_filtered.T,
                                      t_start=lfp.t_start,
                                      sampling_rate=lfp.sampling_rate)
        else:
            output = neo.AnalogSignal(csd_pqarr.T, t_start=lfp.t_start,
                                      sampling_rate=lfp.sampling_rate)
        output.annotate(x_coords=coordinates)
    return output


@deprecated_alias(ele_xx='x_positions', ele_yy='y_positions',
                  ele_zz='z_positions', xlims='x_limits', ylims='y_limits',
                  zlims='z_limits', res='resolution')
def generate_lfp(csd_profile, x_positions, y_positions=None, z_positions=None,
                 x_limits=[0., 1.], y_limits=[0., 1.], z_limits=[0., 1.],
                 resolution=50):
    """
    Forward modelling for getting the potentials for testing Current Source
    Density (CSD).

    Parameters
    ----------
    csd_profile : callable
        A function that computes true CSD profile.
        Available options are (see ./csd/utility_functions.py)
        1D : gauss_1d_dipole
        2D : large_source_2D and small_source_2D
        3D : gauss_3d_dipole
    x_positions : np.ndarray
        A 2D column vector (N x 1 array) containing the positions of the x
        coordinates of the electrodes
    y_positions : np.ndarray, optional
        A 2D column vector (N x 1 array) containing the positions of the y
        coordinates of the electrodes
        Defaults to None, use in 2D or 3D cases only
    z_positions : np.ndarray, optional
        A 2D column vector (N x 1 array) containing the positions of the z
        coordinates of the electrodes
        Defaults to None, use in 3D case only
    x_limits : list, optional
        A list of [start, end].
        The starting spatial coordinate and the ending for integration
        Defaults to [0.,1.]
    y_limits : list, optional
        A list of [start, end].
        The starting spatial coordinate and the ending for integration
        Defaults to [0.,1.], use only in 2D and 3D case
    z_limits : list, optional
        A list of [start, end].
        The starting spatial coordinate and the ending for integration
        Defaults to [0.,1.], use only in 3D case
    resolution : int, optional
        The resolution of the integration
        Defaults to 50

    Returns
    -------
    LFP : :class:`neo.core.AnalogSignal`
       The potentials created by the csd profile at the electrode positions.
       The electrode positions are attached as an annotation named
       'coordinates'.

    Examples
    --------
    >>> import numpy as np
    >>> from elephant.current_source_density import generate_lfp, estimate_csd
    >>> from elephant.current_source_density_src.utility_functions import gauss_1d_dipole  # noqa
    >>> # 1. Define an array xs to x coordinate values with a length of 2304
    >>> xs=np.linspace(0, 10, 2304)

    >>> # 2. Run generate_lfp(gauss_1d_dipole, xs)
    >>> lfp = generate_lfp(gauss_1d_dipole, xs)

    >>> # 3. Run estimate_csd(lfp, method="StandardCSD")
    >>> csd = estimate_csd(lfp, method="StandardCSD")  #doctest: +ELLIPSIS
    discrete ...
    >>> # 4. Print the results
    >>> print(f"LFPs: {lfp}")
    LFPs: [[-0.01483716 -0.01483396 -0.01483075 ...  0.01219233  0.0121911
       0.01218986]] mV
    >>> print(f"CSD estimate: {csd}")  #doctest: +ELLIPSIS
    CSD estimate: [[-1.00025592e-04 -6.06684588e-05  ...
    """

    def integrate_1D(x0, csd_x, csd, h):
        m = np.sqrt((csd_x - x0) ** 2 + h ** 2) - abs(csd_x - x0)
        y = csd * m
        I = simps(y, csd_x)
        return I

    def integrate_2D(x, y, xlin, ylin, csd, h, X, Y):
        x = np.reshape(x, (1, 1, len(x)))
        y = np.reshape(y, (1, 1, len(y)))
        X = np.expand_dims(X, axis=2)
        Y = np.expand_dims(Y, axis=2)
        csd = np.expand_dims(csd, axis=2)
        m = np.sqrt((x - X) ** 2 + (y - Y) ** 2)
        np.clip(m, a_min=0.0000001, a_max=None, out=m)
        y = np.arcsinh(2 * h / m) * csd
        I = simps(y.T, ylin)
        F = simps(I, xlin)
        return F

    def integrate_3D(x, y, z, csd, xlin, ylin, zlin, X, Y, Z):
        m = np.sqrt((x - X) ** 2 + (y - Y) ** 2 + (z - Z) ** 2)
        np.clip(m, a_min=0.0000001, a_max=None, out=m)
        z = csd / m
        Iy = simps(np.transpose(z, (1, 0, 2)), zlin)
        Iy = simps(Iy, ylin)
        F = simps(Iy, xlin)
        return F

    dim = 1
    if z_positions is not None:
        dim = 3
    elif y_positions is not None:
        dim = 2

    x = np.linspace(x_limits[0], x_limits[1], resolution)
    sigma = 1.0
    h = 50.
    if dim == 1:
        # Handle one dimensional case,
        # see https://github.com/NeuralEnsemble/elephant/issues/546
        if len(x_positions.shape) == 1:
            x_positions = np.expand_dims(x_positions, axis=1)
        chrg_x = x
        csd = csd_profile(chrg_x)
        pots = integrate_1D(x_positions, chrg_x, csd, h)
        pots /= 2. * sigma  # eq.: 26 from Potworowski et al
        ele_pos = x_positions
    elif dim == 2:
        y = np.linspace(y_limits[0], y_limits[1], resolution)
        chrg_x = np.expand_dims(x, axis=1)
        chrg_y = np.expand_dims(y, axis=0)
        csd = csd_profile(chrg_x, chrg_y)
        pots = integrate_2D(x_positions, y_positions,
                            x, y,
                            csd, h,
                            chrg_x, chrg_y)
        pots /= 2 * np.pi * sigma
        ele_pos = np.vstack((x_positions, y_positions)).T
    elif dim == 3:
        y = np.linspace(y_limits[0], y_limits[1], resolution)
        z = np.linspace(z_limits[0], z_limits[1], resolution)
        chrg_x, chrg_y, chrg_z = np.mgrid[
            x_limits[0]: x_limits[1]: complex(0, resolution),
            y_limits[0]: y_limits[1]: complex(0, resolution),
            z_limits[0]: z_limits[1]: complex(0, resolution)
        ]

        csd = csd_profile(chrg_x, chrg_y, chrg_z)

        pots = np.zeros(len(x_positions))
        for ii in range(len(x_positions)):
            pots[ii] = integrate_3D(x_positions[ii], y_positions[ii],
                                    z_positions[ii],
                                    csd,
                                    x, y, z,
                                    chrg_x, chrg_y, chrg_z)
        pots /= 4 * np.pi * sigma
        ele_pos = np.vstack((x_positions, y_positions, z_positions)).T
    ele_pos = ele_pos * pq.mm

    asig = neo.AnalogSignal(np.expand_dims(pots, axis=0),
                            sampling_rate=pq.kHz, units='mV')
    asig.annotate(coordinates=ele_pos)
    return asig



# -*- coding: utf-8 -*-
"""
Unit tests for the ASSET analysis.

:copyright: Copyright 2014-2023 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import itertools
import os
import random
import unittest
import warnings
import tempfile
from pathlib import Path

import neo
import numpy as np
import quantities as pq
import scipy.spatial
from numpy.testing import assert_array_almost_equal, assert_array_equal

from elephant import statistics, kernels
from elephant.spike_train_generation import homogeneous_poisson_process

try:
    import sklearn
except ImportError:
    HAVE_SKLEARN = False
else:
    import elephant.asset.asset as asset

    HAVE_SKLEARN = True
    stretchedmetric2d = asset._stretched_metric_2d

try:
    import pyopencl
    HAVE_PYOPENCL = asset.get_opencl_capability()
except ImportError:
    HAVE_PYOPENCL = False

try:
    import pycuda
    HAVE_CUDA = asset.get_cuda_capability_major() > 0
except ImportError:
    HAVE_CUDA = False


class AssetBinningTestCase(unittest.TestCase):

    def setUp(self):
        spiketrain_1 = neo.SpikeTrain(
            [1.3, 2.1, 3.9999999999, 4.9999], units=pq.ms, t_stop=6*pq.ms)

        spiketrain_2 = neo.SpikeTrain(
            [0.9999999999, 1.9999, 4, 5], units=pq.ms, t_stop=6*pq.ms)

        self.spiketrains_i = [spiketrain_1, spiketrain_2]
        self.spiketrains_j = [spiketrain_2, spiketrain_1]

    def test_bin_tolerance_default(self):
        asset_obj = asset.ASSET(spiketrains_i=self.spiketrains_i,
                                spiketrains_j=self.spiketrains_j,
                                bin_size=1*pq.ms)
        bins_i = asset_obj.spiketrains_binned_i.to_array()
        bins_j = asset_obj.spiketrains_binned_j.to_array()

        # Should shift spikes closer than 1e-8 to the right bin edge.
        # This is the current default tolerance for `BinnedSpikeTrain`.
        expected_bins_i = np.array(
            [[0, 1, 1, 0, 2, 0],
             [0, 2, 0, 0, 1, 1]])
        expected_bins_j = np.array(
            [[0, 2, 0, 0, 1, 1],
             [0, 1, 1, 0, 2, 0]])

        self.assertTrue(np.array_equal(bins_i, expected_bins_i))
        self.assertTrue(np.array_equal(bins_j, expected_bins_j))

    def test_bin_tolerance_none(self):
        asset_obj = asset.ASSET(spiketrains_i=self.spiketrains_i,
                                spiketrains_j=self.spiketrains_j,
                                bin_size=1*pq.ms,
                                bin_tolerance=None)
        bins_i = asset_obj.spiketrains_binned_i.to_array()
        bins_j = asset_obj.spiketrains_binned_j.to_array()

        # Should not shift any spikes. Bin should be the same as the integer
        # part of the time.
        expected_bins_i = np.array(
            [[0, 1, 1, 1, 1, 0],
             [1, 1, 0, 0, 1, 1]])
        expected_bins_j = np.array(
            [[1, 1, 0, 0, 1, 1],
             [0, 1, 1, 1, 1, 0]])

        self.assertTrue(np.array_equal(bins_i, expected_bins_i))
        self.assertTrue(np.array_equal(bins_j, expected_bins_j))

    def test_bin_tolerance_float(self):
        asset_obj = asset.ASSET(spiketrains_i=self.spiketrains_i,
                                spiketrains_j=self.spiketrains_j,
                                bin_size=1*pq.ms,
                                bin_tolerance=1e-3)
        bins_i = asset_obj.spiketrains_binned_i.to_array()
        bins_j = asset_obj.spiketrains_binned_j.to_array()

        # Should shift spikes closer than 1e-3 to the right bin edge.
        expected_bins_i = np.array(
            [[0, 1, 1, 0, 1, 1],
             [0, 1, 1, 0, 1, 1]])
        expected_bins_j = np.array(
            [[0, 1, 1, 0, 1, 1],
             [0, 1, 1, 0, 1, 1]])

        self.assertTrue(np.array_equal(bins_i, expected_bins_i))
        self.assertTrue(np.array_equal(bins_j, expected_bins_j))


@unittest.skipUnless(HAVE_SKLEARN, 'requires sklearn')
class AssetTestCase(unittest.TestCase):

    def setUp(self):
        os.environ['ELEPHANT_USE_OPENCL'] = '0'

    def test_stretched_metric_2d_size(self):
        nr_points = 4
        x = np.arange(nr_points)
        D = stretchedmetric2d(x, x, stretch=1, ref_angle=45)
        self.assertEqual(D.shape, (nr_points, nr_points))

    def test_stretched_metric_2d_deprecation_warning(self):
        nr_points = 4
        x = np.arange(nr_points)
        with self.assertWarns(DeprecationWarning):
            stretchedmetric2d(x, x, stretch=1, ref_angle=45, verbose=True)

    def test_stretched_metric_2d_correct_stretching(self):
        x = (0, 1, 0)
        y = (0, 0, 1)
        stretch = 10
        ref_angle = 0
        D = stretchedmetric2d(x, y, stretch=stretch, ref_angle=ref_angle)
        self.assertEqual(D[0, 1], 1)
        self.assertEqual(D[0, 2], stretch)

    def test_stretched_metric_2d_symmetric(self):
        x = (1, 2, 2)
        y = (1, 2, 0)
        stretch = 10
        D = stretchedmetric2d(x, y, stretch=stretch, ref_angle=45)
        assert_array_almost_equal(D, D.T, decimal=6)

    def test_stretched_metric_2d_equals_euclidean_if_stretch_1(self):
        x = np.arange(10)
        y = y = x ** 2 - 2 * x - 4
        # compute stretched distance matrix
        stretch = 1
        D = stretchedmetric2d(x, y, stretch=stretch, ref_angle=45)
        # Compute Euclidean distance matrix
        points = np.vstack([x, y]).T
        E = scipy.spatial.distance_matrix(points, points)
        #  assert if D is equal E
        assert_array_almost_equal(D, E, decimal=5)

    def test_get_sse_start_and_end_time_bins(self):
        sse = {(1, 2): set([1, 2, 3]),
               (3, 4): set([5, 6]),
               (6, 7): set([0, 1])}
        start, end = asset.get_sse_start_and_end_time_bins(sse)
        self.assertListEqual(start, [1, 2])
        self.assertListEqual(end, [6, 7])

    def test_get_neurons_in_sse(self):
        sse = {(1, 2): set([1, 2, 3]),
               (3, 4): set([5, 6]),
               (6, 7): set([0, 1])}
        neurons = asset.get_neurons_in_sse(sse)
        self.assertListEqual(neurons, [0, 1, 2, 3, 5, 6])

    def test_sse_difference(self):
        a = {(1, 2): set([1, 2, 3]), (3, 4): set([5, 6]), (6, 7): set([0, 1])}
        b = {(1, 2): set([1, 2, 5]), (5, 6): set([0, 2]), (6, 7): set([0, 1])}
        diff_ab_pixelwise = {(3, 4): set([5, 6])}
        diff_ba_pixelwise = {(5, 6): set([0, 2])}
        diff_ab_linkwise = {(1, 2): set([3]), (3, 4): set([5, 6])}
        diff_ba_linkwise = {(1, 2): set([5]), (5, 6): set([0, 2])}
        self.assertEqual(
            asset.synchronous_events_difference(a, b, 'pixelwise'),
            diff_ab_pixelwise)
        self.assertEqual(
            asset.synchronous_events_difference(b, a, 'pixelwise'),
            diff_ba_pixelwise)
        self.assertEqual(
            asset.synchronous_events_difference(a, b, 'linkwise'),
            diff_ab_linkwise)
        self.assertEqual(
            asset.synchronous_events_difference(b, a, 'linkwise'),
            diff_ba_linkwise)

    def test_sse_intersection(self):
        a = {(1, 2): set([1, 2, 3]), (3, 4): set([5, 6]), (6, 7): set([0, 1])}
        b = {(1, 2): set([1, 2, 5]), (5, 6): set([0, 2]), (6, 7): set([0, 1])}
        inters_ab_pixelwise = {(1, 2): set([1, 2, 3]), (6, 7): set([0, 1])}
        inters_ba_pixelwise = {(1, 2): set([1, 2, 5]), (6, 7): set([0, 1])}
        inters_ab_linkwise = {(1, 2): set([1, 2]), (6, 7): set([0, 1])}
        inters_ba_linkwise = {(1, 2): set([1, 2]), (6, 7): set([0, 1])}
        self.assertEqual(
            asset.synchronous_events_intersection(a, b, 'pixelwise'),
            inters_ab_pixelwise)
        self.assertEqual(
            asset.synchronous_events_intersection(b, a, 'pixelwise'),
            inters_ba_pixelwise)
        self.assertEqual(
            asset.synchronous_events_intersection(a, b, 'linkwise'),
            inters_ab_linkwise)
        self.assertEqual(
            asset.synchronous_events_intersection(b, a, 'linkwise'),
            inters_ba_linkwise)

    def test_sse_relations(self):
        a = {(1, 2): set([1, 2, 3]), (3, 4): set([5, 6]), (6, 7): set([0, 1])}
        b = {(1, 2): set([1, 2, 5]), (5, 6): set([0, 2]), (6, 7): set([0, 1])}
        c = {(5, 6): set([0, 2])}
        d = {(3, 4): set([0, 1]), (5, 6): set([0, 1, 2])}
        self.assertTrue(asset.synchronous_events_identical({}, {}))
        self.assertTrue(asset.synchronous_events_identical(a, a))
        self.assertFalse(asset.synchronous_events_identical(b, c))
        self.assertTrue(asset.synchronous_events_no_overlap(a, c))
        self.assertTrue(asset.synchronous_events_no_overlap(a, d))
        self.assertFalse(asset.synchronous_events_no_overlap(a, b))
        self.assertFalse(asset.synchronous_events_no_overlap({}, {}))
        self.assertTrue(asset.synchronous_events_contained_in(c, b))
        self.assertTrue(asset.synchronous_events_contained_in(c, d))
        self.assertFalse(asset.synchronous_events_contained_in(a, d))
        self.assertFalse(asset.synchronous_events_contained_in(a, b))
        self.assertTrue(asset.synchronous_events_contains_all(b, c))
        self.assertTrue(asset.synchronous_events_contains_all(d, c))
        self.assertFalse(asset.synchronous_events_contains_all(a, b))
        self.assertTrue(asset.synchronous_events_overlap(a, b))
        self.assertFalse(asset.synchronous_events_overlap(c, d))

    def test_mask_matrix(self):
        mat1 = np.array([[0, 1], [1, 2]])
        mat2 = np.array([[2, 1], [1, 3]])

        mask_1_2 = asset.ASSET.mask_matrices([mat1, mat2], [1, 2])
        mask_1_2_correct = np.array([[False, False], [False, True]])
        self.assertTrue(np.all(mask_1_2 == mask_1_2_correct))
        self.assertIsInstance(mask_1_2[0, 0], np.bool_)

        self.assertRaises(ValueError, asset.ASSET.mask_matrices, [], [])
        self.assertRaises(ValueError, asset.ASSET.mask_matrices,
                          [np.arange(5)], [])

    def test_cluster_matrix_entries(self):
        # test with symmetric matrix
        mat = np.array([[0, 0, 1, 0],
                        [0, 0, 0, 1],
                        [1, 0, 0, 0],
                        [0, 1, 0, 0]])

        clustered = asset.ASSET.cluster_matrix_entries(
            mat, max_distance=1.5, min_neighbors=2, stretch=1)
        correct = np.array([[0, 0, 1, 0],
                            [0, 0, 0, 1],
                            [2, 0, 0, 0],
                            [0, 2, 0, 0]])
        assert_array_equal(clustered, correct)

        # test with non-symmetric matrix
        mat = np.array([[0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [1, 0, 0, 1],
                        [0, 1, 0, 0]])
        clustered = asset.ASSET.cluster_matrix_entries(
            mat, max_distance=1.5, min_neighbors=3, stretch=1)
        correct = np.array([[0, 1, 0, 0],
                            [0, 0, 1, 0],
                            [-1, 0, 0, 1],
                            [0, -1, 0, 0]])
        assert_array_equal(clustered, correct)

        # test with lowered min_neighbors
        mat = np.array([[0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [1, 0, 0, 1],
                        [0, 1, 0, 0]])
        clustered = asset.ASSET.cluster_matrix_entries(
            mat, max_distance=1.5, min_neighbors=2, stretch=1)
        correct = np.array([[0, 1, 0, 0],
                            [0, 0, 1, 0],
                            [2, 0, 0, 1],
                            [0, 2, 0, 0]])
        assert_array_equal(clustered, correct)

        mat = np.zeros((4, 4))
        clustered = asset.ASSET.cluster_matrix_entries(
            mat, max_distance=1.5, min_neighbors=2, stretch=1)
        correct = mat
        assert_array_equal(clustered, correct)

    def test_cluster_matrix_entries_chunked(self):
        np.random.seed(12)
        mmat = np.random.randn(50, 50) > 0
        max_distance = 5
        min_neighbors = 4
        stretch = 2
        cmat_true = asset.ASSET.cluster_matrix_entries(
            mmat, max_distance=max_distance, min_neighbors=min_neighbors,
            stretch=stretch)
        for working_memory in [1, 10, 100, 1000]:
            cmat = asset.ASSET.cluster_matrix_entries(
                mmat, max_distance=max_distance, min_neighbors=min_neighbors,
                stretch=stretch, working_memory=working_memory)
            assert_array_equal(cmat, cmat_true)

    def test_cluster_matrix_entries_chunked_array_file(self):
        np.random.seed(12)
        mmat = np.random.randn(100, 100) > 0
        max_distance = 2
        min_neighbors = 2
        stretch = 2
        cmat_true = asset.ASSET.cluster_matrix_entries(
            mmat, max_distance=max_distance, min_neighbors=min_neighbors,
            stretch=stretch)

        for working_memory in [1, 10, 100, 1000]:
            with tempfile.TemporaryDirectory() as tmpdir:
                cmat = asset.ASSET.cluster_matrix_entries(
                    mmat, max_distance=max_distance,
                    min_neighbors=min_neighbors, stretch=stretch,
                    working_memory=working_memory,
                    array_file=Path(tmpdir) / f"test_dist_{working_memory}")
                assert_array_equal(cmat, cmat_true)

    def test_pmat_neighbors_gpu(self):
        np.random.seed(12)
        n_largest = 3
        pmat1 = np.random.random_sample((40, 40)).astype(np.float32)
        np.fill_diagonal(pmat1, 0.5)
        pmat2 = np.random.random_sample((70, 23)).astype(np.float32)
        pmat3 = np.random.random_sample((27, 93)).astype(np.float32)
        for pmat in (pmat1, pmat2, pmat3):
            for filter_size in (4, 8, 11):
                filter_shape = (filter_size, 3)
                with warnings.catch_warnings():
                    # ignore even filter sizes
                    warnings.simplefilter('ignore', UserWarning)
                    pmat_neigh = asset._PMatNeighbors(
                        filter_shape=filter_shape, n_largest=n_largest)
                lmat_true = pmat_neigh.cpu(pmat)
                if HAVE_PYOPENCL:
                    lmat_opencl = pmat_neigh.pyopencl(pmat)
                    assert_array_almost_equal(lmat_opencl, lmat_true)
                if HAVE_CUDA:
                    lmat_cuda = pmat_neigh.pycuda(pmat)
                    assert_array_almost_equal(lmat_cuda, lmat_true)

    def test_pmat_neighbors_gpu_chunked(self):
        np.random.seed(12)
        filter_shape = (7, 3)
        n_largest = 3
        pmat1 = np.random.random_sample((40, 40)).astype(np.float32)
        np.fill_diagonal(pmat1, 0.5)
        pmat2 = np.random.random_sample((70, 27)).astype(np.float32)
        pmat3 = np.random.random_sample((41, 80)).astype(np.float32)
        for pmat in (pmat1, pmat2, pmat3):
            pmat_neigh = asset._PMatNeighbors(filter_shape=filter_shape,
                                              n_largest=n_largest)
            lmat_true = pmat_neigh.cpu(pmat)
            for max_chunk_size in (17, 20, 29):
                pmat_neigh.max_chunk_size = max_chunk_size
                if HAVE_PYOPENCL:
                    lmat_opencl = pmat_neigh.pyopencl(pmat)
                    assert_array_almost_equal(lmat_opencl, lmat_true)
                if HAVE_CUDA:
                    lmat_cuda = pmat_neigh.pycuda(pmat)
                    assert_array_almost_equal(lmat_cuda, lmat_true)

    def test_pmat_neighbors_gpu_overlapped_chunks(self):
        # The pmat is chunked as follows:
        #   [(0, 11), (11, 22), (22, 33), (29, 40)]
        # Two last chunks overlap.
        np.random.seed(12)
        pmat = np.random.random_sample((50, 50)).astype(np.float32)
        pmat_neigh = asset._PMatNeighbors(filter_shape=(11, 5), n_largest=3,
                                          max_chunk_size=12)
        lmat_true = pmat_neigh.cpu(pmat)
        if HAVE_PYOPENCL:
            lmat_opencl = pmat_neigh.pyopencl(pmat)
            assert_array_almost_equal(lmat_opencl, lmat_true)
        if HAVE_CUDA:
            lmat_cuda = pmat_neigh.pycuda(pmat)
            assert_array_almost_equal(lmat_cuda, lmat_true)

    def test_pmat_neighbors_deprecation_warning(self):
        with self.assertWarns(DeprecationWarning):
            asset._PMatNeighbors(filter_shape=(11, 5), n_largest=3,
                                 max_chunk_size=12, verbose=True)

    def test_pmat_neighbors_invalid_input(self):
        np.random.seed(12)
        pmat = np.random.random_sample((20, 20))
        np.fill_diagonal(pmat, 0.5)

        # Too large filter_shape
        pmat_neigh = asset._PMatNeighbors(filter_shape=(11, 3), n_largest=3)
        self.assertRaises(ValueError, pmat_neigh.compute, pmat)
        pmat_neigh = asset._PMatNeighbors(filter_shape=(21, 3), n_largest=3)
        np.fill_diagonal(pmat, 0.0)
        self.assertRaises(ValueError, pmat_neigh.compute, pmat)
        pmat_neigh = asset._PMatNeighbors(filter_shape=(11, 3), n_largest=3,
                                          max_chunk_size=10)
        if HAVE_PYOPENCL:
            # max_chunk_size > filter_shape
            self.assertRaises(ValueError, pmat_neigh.pyopencl, pmat)
        if HAVE_CUDA:
            # max_chunk_size > filter_shape
            self.assertRaises(ValueError, pmat_neigh.pycuda, pmat)

        # Too small filter_shape
        self.assertRaises(ValueError, asset._PMatNeighbors,
                          filter_shape=(11, 3), n_largest=100)

        # w >= l
        self.assertRaises(ValueError, asset._PMatNeighbors,
                          filter_shape=(9, 9), n_largest=3)

        # not centered
        self.assertWarns(UserWarning, asset._PMatNeighbors,
                         filter_shape=(10, 6), n_largest=3)

    def test_intersection_matrix(self):
        st1 = neo.SpikeTrain([1, 2, 4] * pq.ms, t_stop=6 * pq.ms)
        st2 = neo.SpikeTrain([1, 3, 4] * pq.ms, t_stop=6 * pq.ms)
        st3 = neo.SpikeTrain([2, 5] * pq.ms, t_start=1 * pq.ms,
                             t_stop=6 * pq.ms)
        bin_size = 1 * pq.ms

        asset_obj_same_t_start_stop = asset.ASSET(
            [st1, st2], bin_size=bin_size, t_stop_i=5 * pq.ms,
            t_stop_j=5 * pq.ms)

        # Check that the routine works for correct input...
        # ...same t_start, t_stop on both time axes
        imat_1_2 = asset_obj_same_t_start_stop.intersection_matrix()
        trueimat_1_2 = np.array([[0., 0., 0., 0., 0.],
                                 [0., 2., 1., 1., 2.],
                                 [0., 1., 1., 0., 1.],
                                 [0., 1., 0., 1., 1.],
                                 [0., 2., 1., 1., 2.]])
        assert_array_equal(asset_obj_same_t_start_stop.x_edges,
                           np.arange(6) * pq.ms)  # correct bins
        assert_array_equal(asset_obj_same_t_start_stop.y_edges,
                           np.arange(6) * pq.ms)  # correct bins
        assert_array_equal(imat_1_2, trueimat_1_2)  # correct matrix
        # ...different t_start, t_stop on the two time axes
        asset_obj_different_t_start_stop = asset.ASSET(
            [st1, st2], spiketrains_j=[st + 6 * pq.ms for st in [st1, st2]],
            bin_size=bin_size, t_start_j=6 * pq.ms, t_stop_i=5 * pq.ms,
            t_stop_j=11 * pq.ms)
        imat_1_2 = asset_obj_different_t_start_stop.intersection_matrix()
        assert_array_equal(asset_obj_different_t_start_stop.x_edges,
                           np.arange(6) * pq.ms)  # correct bins
        assert_array_equal(asset_obj_different_t_start_stop.y_edges,
                           np.arange(6, 12) * pq.ms)
        self.assertTrue(np.all(imat_1_2 == trueimat_1_2))  # correct matrix

        # test with norm=1
        imat_1_2 = asset_obj_same_t_start_stop.intersection_matrix(
            normalization='intersection')
        trueimat_1_2 = np.array([[0., 0., 0., 0., 0.],
                                 [0., 1., 1., 1., 1.],
                                 [0., 1., 1., 0., 1.],
                                 [0., 1., 0., 1., 1.],
                                 [0., 1., 1., 1., 1.]])
        assert_array_equal(imat_1_2, trueimat_1_2)

        # test with norm=2
        imat_1_2 = asset_obj_same_t_start_stop.intersection_matrix(
            normalization='mean')
        sq = np.sqrt(2) / 2.
        trueimat_1_2 = np.array([[0., 0., 0., 0., 0.],
                                 [0., 1., sq, sq, 1.],
                                 [0., sq, 1., 0., sq],
                                 [0., sq, 0., 1., sq],
                                 [0., 1., sq, sq, 1.]])
        assert_array_almost_equal(imat_1_2, trueimat_1_2)

        # test with norm=3
        imat_1_2 = asset_obj_same_t_start_stop.intersection_matrix(
            normalization='union')
        trueimat_1_2 = np.array([[0., 0., 0., 0., 0.],
                                 [0., 1., .5, .5, 1.],
                                 [0., .5, 1., 0., .5],
                                 [0., .5, 0., 1., .5],
                                 [0., 1., .5, .5, 1.]])
        assert_array_almost_equal(imat_1_2, trueimat_1_2)

        # Check that errors are raised correctly...
        # ...for partially overlapping time intervals
        self.assertRaises(ValueError, asset.ASSET,
                          spiketrains_i=[st1, st2], bin_size=bin_size,
                          t_start_j=1 * pq.ms)
        # ...for different SpikeTrain's t_starts
        self.assertRaises(ValueError, asset.ASSET,
                          spiketrains_i=[st1, st3], bin_size=bin_size)
        # ...for different SpikeTrain's t_stops
        self.assertRaises(ValueError, asset.ASSET,
                          spiketrains_i=[st1, st2], bin_size=bin_size,
                          t_stop_j=5 * pq.ms)

    #  regression test Issue #481
    #  see: https://github.com/NeuralEnsemble/elephant/issues/481
    def test_asset_choose_backend_opencl(self):
        class TestClassBackend(asset._GPUBackend):

            def __init__(self):
                super().__init__()
                self.backend = self._choose_backend()

            def cpu(self):
                return "cpu"

            def pycuda(self):
                return "cuda"

            def pyopencl(self):
                return "opencl"

        # check which backend is chosen if environment variable for opencl
        # is not set
        os.environ.pop('ELEPHANT_USE_OPENCL', None)
        # create object of TestClass
        backend_obj = TestClassBackend()

        if HAVE_PYOPENCL:
            self.assertEqual(backend_obj.backend(), 'opencl')
        else:
            # if environment variable is not set and no module pyopencl or
            # device is found: choose cpu backend
            self.assertEqual(backend_obj.backend(), 'cpu')

        os.environ['ELEPHANT_USE_OPENCL'] = '0'

    def test_asset_deprecation_warning(self):
        st1 = neo.SpikeTrain([1, 2, 4] * pq.ms, t_stop=6 * pq.ms)
        st2 = neo.SpikeTrain([1, 3, 4] * pq.ms, t_stop=6 * pq.ms)
        bin_size = 1 * pq.ms
        with self.assertWarns(DeprecationWarning):
            asset.ASSET([st1, st2], bin_size=bin_size, t_stop_i=5 * pq.ms,
                        t_stop_j=5 * pq.ms, verbose=True)


@unittest.skipUnless(HAVE_SKLEARN, 'requires sklearn')
class TestJSFUniformOrderStat3D(unittest.TestCase):

    def test_combinations_with_replacement(self):
        # Test that _combinations_with_replacement yields the same tuples
        # as in the original implementation with itertools.product(*lists)
        # and filtering by _wrong_order.

        def _wrong_order(a):
            if a[-1] > a[0]:
                return True
            for i in range(len(a) - 1):
                if a[i] < a[i + 1]:
                    return True
            return False

        for n in range(1, 15):
            for d in range(1, min(6, n + 1)):
                jsf = asset._JSFUniformOrderStat3D(n=n, d=d)
                lists = [range(j, n + 1) for j in range(d, 0, -1)]
                matrix_entries = list(
                    jsf._combinations_with_replacement()
                )
                matrix_entries_correct = [
                    indices for indices in itertools.product(*lists)
                    if not _wrong_order(indices)
                ]
                self.assertEqual(matrix_entries, matrix_entries_correct)
                self.assertEqual(jsf.num_iterations,
                                 len(matrix_entries_correct))

    def test_next_sequence_sorted(self):
        # This test shows the main idea of CUDA ASSET parallelization that
        # allows reconstructing a 'sequence_sorted' from the iteration number
        # and therefore getting rid of sequential nature of joint prob.
        # matrix calculation.
        def next_sequence_sorted(jsf, iteration):
            # One-to-one correspondence with
            # 'void next_sequence_sorted(int *sequence_sorted, ULL iteration)'
            # function in asset.pycuda.py.
            sequence_sorted = []
            element = jsf.n - 1
            for row in range(jsf.d - 1, -1, -1):
                map_row = jsf.map_iterations[row]
                while element > row and iteration < map_row[element]:
                    element -= 1
                iteration -= map_row[element]
                sequence_sorted.append(element + 1)
            return tuple(sequence_sorted)

        for n in range(1, 15):
            for d in range(1, min(6, n + 1)):
                jsf = asset._JSFUniformOrderStat3D(n=n, d=d)
                for iter_id, seq_sorted_true in enumerate(
                        jsf._combinations_with_replacement()):
                    seq_sorted = next_sequence_sorted(jsf, iteration=iter_id)
                    self.assertEqual(seq_sorted, seq_sorted_true)

    def test_invalid_values(self):
        # 1) d > n
        self.assertRaises(ValueError, asset._JSFUniformOrderStat3D, n=5, d=6)

        # 2) d does not match the input data shape
        d = 4
        jsf = asset._JSFUniformOrderStat3D(n=5, d=d)
        u = np.empty((3, d + 1))
        self.assertRaises(ValueError, jsf.compute, u=u)

    def test_JSFUniformOrderStat3D_deprecation_warning(self):
        with self.assertWarns(DeprecationWarning):
            asset._JSFUniformOrderStat3D(n=5, d=1, verbose=True)

    def test_point_mass_output(self):
        # When N >> D, the expected output is [1, 0]
        L, N, D = 2, 50, 2
        jsf = asset._JSFUniformOrderStat3D(n=N, d=D, precision='double')
        u = np.arange(L * D, dtype=np.float32).reshape((-1, D))
        u /= np.max(u)
        p_out = jsf.compute(u)
        assert_array_almost_equal(p_out, [1., 0.], decimal=5)

    def test_precision(self):
        L = 2
        for n in range(1, 10):
            for d in range(1, min(6, n + 1)):
                u = np.arange(L * d, dtype=np.float32).reshape((-1, d))
                u /= np.max(u)
                jsf_double = asset._JSFUniformOrderStat3D(n=n, d=d,
                                                          precision='double')
                jsf_float = asset._JSFUniformOrderStat3D(n=n, d=d,
                                                         precision='float')
                P_total_double = jsf_double.compute(u)
                P_total_float = jsf_float.compute(u)
                # decimal 5 is used because the number of iterations is small
                # in practice, the results deviate starting at decimal 3 or 2
                assert_array_almost_equal(P_total_double, P_total_float,
                                          decimal=5)

    def test_gpu(self):
        np.random.seed(12)
        for precision, L, n in itertools.product(['float', 'double'],
                                                 [1, 23, 100], [1, 3, 10]):
            for d in range(1, min(4, n + 1)):
                u = np.random.rand(L, d).cumsum(axis=1)
                u /= np.max(u)
                jsf = asset._JSFUniformOrderStat3D(n=n, d=d,
                                                   precision=precision)
                du = np.diff(u, prepend=0, append=1, axis=1)
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', RuntimeWarning)
                    log_du = np.log(du, dtype=np.float32)
                P_total_cpu = jsf.cpu(log_du)
                for max_chunk_size in [None, 22]:
                    jsf.max_chunk_size = max_chunk_size
                    if HAVE_PYOPENCL:
                        P_total_opencl = jsf.pyopencl(log_du)
                        assert_array_almost_equal(P_total_opencl, P_total_cpu)
                    if HAVE_CUDA:
                        P_total_cuda = jsf.pycuda(log_du)
                        assert_array_almost_equal(P_total_cuda, P_total_cpu)

    def test_gpu_threads_and_cwr_loops(self):
        # The num. of threads (work items) and CWR loops must not influence
        # the result.
        L, N, D = 100, 10, 3

        u = np.arange(L * D, dtype=np.float32).reshape((-1, D))
        u /= np.max(u)
        du = np.diff(u, prepend=0, append=1, axis=1)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            log_du = np.log(du, dtype=np.float32)

        def run_test(jsf, jsf_backend):
            P_expected = jsf_backend(log_du)
            for threads in [1, 16, 32, 128, 1024]:
                for cwr_loops in [1, 16, 128]:
                    jsf.cuda_threads = threads
                    jsf.cuda_cwr_loops = cwr_loops
                    P_total = jsf_backend(log_du)
                    assert_array_almost_equal(P_total, P_expected)

        if HAVE_PYOPENCL:
            jsf = asset._JSFUniformOrderStat3D(n=N, d=D, precision='float')
            run_test(jsf, jsf.pyopencl)
        if HAVE_CUDA:
            jsf = asset._JSFUniformOrderStat3D(n=N, d=D, precision='float')
            run_test(jsf, jsf.pycuda)

    def test_gpu_chunked(self):
        L, N, D = 100, 9, 3
        u = np.arange(L * D, dtype=np.float32).reshape((-1, D))
        u /= np.max(u)
        du = np.diff(u, prepend=0, append=1, axis=1)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            log_du = np.log(du, dtype=np.float32)
        jsf = asset._JSFUniformOrderStat3D(n=N, d=D)
        P_true = jsf.cpu(log_du)
        for max_chunk_size in (13, 50):
            jsf.max_chunk_size = max_chunk_size
            if HAVE_PYOPENCL:
                P_total = jsf.pyopencl(log_du)
                assert_array_almost_equal(P_total, P_true)
            if HAVE_CUDA:
                P_total = jsf.pycuda(log_du)
                assert_array_almost_equal(P_total, P_true)

    def test_watchdog(self):
        L, N, D = 10, 7, 3
        u = np.arange(L * D, dtype=np.float32).reshape((-1, D))
        u /= np.max(u) - 1
        # 'u' is invalid input, which must lead to watchdog barking
        jsf = asset._JSFUniformOrderStat3D(n=N, d=D)
        self.assertWarns(UserWarning, jsf.compute, u)


@unittest.skipUnless(HAVE_SKLEARN, 'requires sklearn')
class AssetTestIntegration(unittest.TestCase):
    def setUp(self):
        # common for all tests
        self.bin_size = 3 * pq.ms

    def test_probability_matrix_symmetric(self):
        np.random.seed(1)
        kernel_width = 9 * pq.ms
        rate = 50 * pq.Hz
        n_spiketrains = 50
        spiketrains = []
        spiketrains_copy = []
        for _ in range(n_spiketrains):
            st = homogeneous_poisson_process(rate, t_stop=100 * pq.ms)
            spiketrains.append(st)
            spiketrains_copy.append(st.copy())

        asset_obj = asset.ASSET(spiketrains, bin_size=self.bin_size)
        asset_obj_symmetric = asset.ASSET(spiketrains,
                                          spiketrains_j=spiketrains_copy,
                                          bin_size=self.bin_size)

        imat = asset_obj.intersection_matrix()
        pmat = asset_obj.probability_matrix_analytical(
            kernel_width=kernel_width)

        imat_symm = asset_obj_symmetric.intersection_matrix()
        pmat_symm = asset_obj_symmetric.probability_matrix_analytical(
            kernel_width=kernel_width)

        assert_array_almost_equal(pmat, pmat_symm)
        assert_array_almost_equal(imat, imat_symm)
        assert_array_almost_equal(asset_obj.x_edges,
                                  asset_obj_symmetric.x_edges)
        assert_array_almost_equal(asset_obj.y_edges,
                                  asset_obj_symmetric.y_edges)

    def _test_integration_subtest(self, spiketrains, spiketrains_y,
                                  indices_pmat, index_proba, expected_sses):
        # define parameters
        random.seed(1)
        kernel_width = 9 * pq.ms
        surrogate_dt = 9 * pq.ms
        alpha = 0.9
        filter_shape = (5, 1)
        nr_largest = 3
        max_distance = 3
        min_neighbors = 3
        stretch = 5
        n_surr = 20

        def _get_rates(_spiketrains):
            kernel_sigma = kernel_width / 2. / np.sqrt(3.)
            kernel = kernels.RectangularKernel(sigma=kernel_sigma)
            rates = [statistics.instantaneous_rate(
                st,
                kernel=kernel,
                sampling_period=1 * pq.ms)
                for st in _spiketrains]
            return rates

        asset_obj = asset.ASSET(spiketrains, spiketrains_y,
                                bin_size=self.bin_size)

        # calculate the intersection matrix
        imat = asset_obj.intersection_matrix()

        # calculate probability matrix analytical
        pmat = asset_obj.probability_matrix_analytical(
            imat,
            kernel_width=kernel_width)

        # check if pmat is the same when rates are provided
        pmat_as_rates = asset_obj.probability_matrix_analytical(
            imat,
            firing_rates_x=_get_rates(spiketrains),
            firing_rates_y=_get_rates(spiketrains_y))
        assert_array_almost_equal(pmat, pmat_as_rates)

        # calculate probability matrix montecarlo
        pmat_montecarlo = asset_obj.probability_matrix_montecarlo(
            n_surrogates=n_surr,
            imat=imat,
            surrogate_dt=surrogate_dt,
            surrogate_method='dither_spikes')

        # test probability matrices
        assert_array_equal(np.where(pmat > alpha), indices_pmat)
        assert_array_equal(np.where(pmat_montecarlo > alpha),
                           indices_pmat)
        # calculate joint probability matrix
        jmat = asset_obj.joint_probability_matrix(
            pmat,
            filter_shape=filter_shape,
            n_largest=nr_largest)
        # test joint probability matrix
        assert_array_equal(np.where(jmat > 0.98), index_proba['high'])
        assert_array_equal(np.where(jmat > 0.9), index_proba['medium'])
        assert_array_equal(np.where(jmat > 0.8), index_proba['low'])
        # test if all other entries are zeros
        mask_zeros = np.ones(jmat.shape, bool)
        mask_zeros[index_proba['low']] = False
        self.assertTrue(np.all(jmat[mask_zeros] == 0))

        # calculate mask matrix and cluster matrix
        mmat = asset_obj.mask_matrices([pmat, jmat], [alpha, alpha])
        cmat = asset_obj.cluster_matrix_entries(
            mmat,
            max_distance=max_distance,
            min_neighbors=min_neighbors,
            stretch=stretch)

        # extract sses and test them
        sses = asset_obj.extract_synchronous_events(cmat)
        self.assertDictEqual(sses, expected_sses)

    def test_integration(self):
        """
        The test is written according to the notebook (for developers only):
        https://github.com/INM-6/elephant-tutorials/blob/master/
        simple_test_asset.ipynb
        """
        # define parameters
        np.random.seed(1)
        size_group = 3
        size_sse = 3
        T = 60 * pq.ms
        delay = 9 * pq.ms
        bins_between_sses = 3
        time_between_sses = 9 * pq.ms
        # ground truth for pmats
        starting_bin_1 = int((delay / self.bin_size).magnitude.item())
        starting_bin_2 = int(
            (2 * delay / self.bin_size + time_between_sses / self.bin_size
             ).magnitude.item())
        indices_pmat_1 = np.arange(starting_bin_1, starting_bin_1 + size_sse)
        indices_pmat_2 = np.arange(starting_bin_2,
                                   starting_bin_2 + size_sse)
        indices_pmat = (np.concatenate((indices_pmat_1, indices_pmat_2)),
                        np.concatenate((indices_pmat_2, indices_pmat_1)))
        # generate spike trains
        spiketrains = [neo.SpikeTrain([index_spiketrain,
                                       index_spiketrain +
                                       size_sse +
                                       bins_between_sses] * self.bin_size
                                      + delay + 1 * pq.ms,
                                      t_stop=T)
                       for index_group in range(size_group)
                       for index_spiketrain in range(size_sse)]
        index_proba = {
            "high": (np.array([9, 9, 10, 10, 10, 11, 11]),
                     np.array([3, 4, 3, 4, 5, 4, 5])),
            "medium": (np.array([8, 8, 9, 9, 9, 10, 10,
                                 10, 11, 11, 11, 12, 12]),
                       np.array([2, 3, 2, 3, 4, 3, 4, 5, 4, 5, 6, 5, 6])),
            "low": (np.array([7, 8, 8, 9, 9, 9, 10, 10, 10,
                              11, 11, 11, 12, 12, 12, 13, 13]),
                    np.array([2, 2, 3, 2, 3, 4, 3, 4, 5, 4, 5,
                              6, 5, 6, 7, 6, 7]))
        }
        expected_sses = {1: {(9, 3): {0, 3, 6}, (10, 4): {1, 4, 7},
                             (11, 5): {2, 5, 8}}}
        self._test_integration_subtest(spiketrains,
                                       spiketrains_y=spiketrains,
                                       indices_pmat=indices_pmat,
                                       index_proba=index_proba,
                                       expected_sses=expected_sses)

    def test_integration_nonsymmetric(self):
        # define parameters
        np.random.seed(1)
        random.seed(1)
        size_group = 3
        size_sse = 3
        delay = 18 * pq.ms
        T = 4 * delay + 2 * size_sse * self.bin_size
        time_between_sses = 2 * delay
        # ground truth for pmats
        starting_bin = int((delay / self.bin_size).magnitude.item())
        indices_pmat_1 = np.arange(starting_bin, starting_bin + size_sse)
        indices_pmat = (indices_pmat_1, indices_pmat_1)
        # generate spike trains
        spiketrains = [
            neo.SpikeTrain([index_spiketrain] * self.bin_size + delay,
                           t_start=0 * pq.ms,
                           t_stop=2 * delay + size_sse * self.bin_size)
            for index_group in range(size_group)
            for index_spiketrain in range(size_sse)]
        spiketrains_y = [
            neo.SpikeTrain([index_spiketrain] * self.bin_size + delay +
                           time_between_sses + size_sse * self.bin_size,
                           t_start=size_sse * self.bin_size + 2 * delay,
                           t_stop=T)
            for index_group in range(size_group)
            for index_spiketrain in range(size_sse)]
        index_proba = {
            "high": ([6, 6, 7, 7, 7, 8, 8],
                     [6, 7, 6, 7, 8, 7, 8]),
            "medium": ([5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9],
                       [5, 6, 5, 6, 7, 6, 7, 8, 7, 8, 9, 8, 9]),
            "low": ([4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7,
                     8, 8, 8, 9, 9, 9, 10, 10],
                    [4, 5, 4, 5, 6, 5, 6, 7, 6, 7, 8,
                     7, 8, 9, 8, 9, 10, 9, 10])
        }
        expected_sses = {1: {(6, 6): {0, 3, 6}, (7, 7): {1, 4, 7},
                             (8, 8): {2, 5, 8}}}
        self._test_integration_subtest(spiketrains,
                                       spiketrains_y=spiketrains_y,
                                       indices_pmat=indices_pmat,
                                       index_proba=index_proba,
                                       expected_sses=expected_sses)


if __name__ == "__main__":
    unittest.main()
